from unittest import TestCase
from unittest.mock import MagicMock, patch

from piccolo.apps.sql_shell.commands.run import run
from tests.base import postgres_only, sqlite_only


class TestRun(TestCase):
    @postgres_only
    @patch("piccolo.apps.sql_shell.commands.run.subprocess")
    def test_psql(self, subprocess: MagicMock):
        """
        Make sure psql was called correctly.
        """
        run()
        self.assertTrue(subprocess.run.called)

        assert subprocess.run.call_args.args[0] == [
            "psql",
            "-U",
            "postgres",
            "-h",
            "localhost",
            "-p",
            "5432",
            "piccolo",
        ]

    @sqlite_only
    @patch("piccolo.apps.sql_shell.commands.run.subprocess")
    def test_sqlite3(self, subprocess: MagicMock):
        """
        Make sure sqlite3 was called correctly.
        """
        run()
        self.assertTrue(subprocess.run.called)

        assert subprocess.run.call_args.args[0] == ["sqlite3", "test.sqlite"]



from piccolo.columns import Integer
from piccolo.query.functions.conditional import Coalesce
from piccolo.table import Table
from piccolo.testing.test_case import AsyncTableTest


class Band(Table):
    popularity = Integer(null=True, default=None)


class TestCoalesce(AsyncTableTest):

    tables = [Band]

    async def asyncSetUp(self):
        await super().asyncSetUp()
        await Band({Band.popularity: None}).save()

    async def test_coalesce(self):
        response = await Band.select(Coalesce(Band.popularity, 10))
        self.assertListEqual(response, [{"popularity": 10}])

    async def test_coalesce_pipe_syntax(self):
        response = await Band.select(Band.popularity | 10)
        self.assertListEqual(response, [{"popularity": 10}])



import timeit
from dataclasses import dataclass
from typing import Any, Union
from unittest import mock

from piccolo.columns import Integer, Varchar
from piccolo.query.base import FrozenQuery, Query
from piccolo.table import Table
from tests.base import AsyncMock, DBTestCase, sqlite_only
from tests.example_apps.music.tables import Band


@dataclass
class QueryResponse:
    query: Union[Query, FrozenQuery]
    response: Any


class TestFreeze(DBTestCase):
    def test_frozen_select_queries(self) -> None:
        """
        Make sure a variety of select queries work as expected when frozen.
        """
        self.insert_rows()

        query_responses: list[QueryResponse] = [
            QueryResponse(
                query=(
                    Band.select(Band.name)
                    .order_by(Band.popularity, ascending=False)
                    .first()
                    .freeze()
                ),
                response={"name": "Rustaceans"},
            ),
            QueryResponse(
                query=(
                    Band.select(Band.name)
                    .order_by(Band.popularity, ascending=False)
                    .freeze()
                ),
                response=[
                    {"name": "Rustaceans"},
                    {"name": "Pythonistas"},
                    {"name": "CSharps"},
                ],
            ),
            QueryResponse(
                query=(
                    Band.select(Band.name)
                    .where(Band.name == "Pythonistas")
                    .freeze()
                ),
                response=[{"name": "Pythonistas"}],
            ),
        ]

        for query_response in query_responses:
            result = query_response.query.run_sync()
            self.assertEqual(result, query_response.response)

    def test_output_clause(self):
        """
        Make sure the output clause still works correctly with frozen queries.
        """
        self.insert_rows()

        result = (
            Band.select(Band.name)
            .where(Band.name == "Pythonistas")
            .output(as_json=True)
            .freeze()
            .run_sync()
        )
        # Some JSON encoders have a space after the colons and some don't,
        # so normalise them.
        self.assertEqual(result.replace(" ", ""), '[{"name":"Pythonistas"}]')

    @sqlite_only
    def test_frozen_performance(self):
        """
        The frozen query performance should exceed the non-frozen. If not,
        there's a problem.

        We mock out the database to make the performance more predictable.

        """
        db = mock.MagicMock()
        db.engine_type = "sqlite"
        db.run_querystring = AsyncMock()
        db.run_querystring.return_value = [
            {"name": "Pythonistas", "popularity": 1000}
        ]

        class Band(Table, db=db):
            name = Varchar()
            popularity = Integer()

        iterations = 50
        query = (
            Band.select(Band.name)
            .where(Band.popularity > 900)
            .order_by(Band.name)
        )
        query_duration = timeit.repeat(
            lambda: query.run_sync(), repeat=iterations, number=1
        )

        frozen_query = query.freeze()
        frozen_query_duration = timeit.repeat(
            lambda: frozen_query.run_sync(), repeat=iterations, number=1
        )

        # Remove the outliers before comparing
        self.assertGreater(
            sum(sorted(query_duration)[10:-10]),
            sum(sorted(frozen_query_duration)[10:-10]),
        )

    def test_attribute_access(self):
        """
        Once frozen, you shouldn't be able to call additional methods on the
        query (for example `.where`).
        """
        query = Band.select().freeze()

        with self.assertRaises(AttributeError):
            query.where(Band.name == "Pythonistas")



from .base import *  # noqa
from .date import *  # noqa
from .interval import *  # noqa
from .time import *  # noqa
from .timestamp import *  # noqa
from .timestamptz import *  # noqa
from .uuid import *  # noqa



import decimal
import uuid

import pytest

from piccolo.columns.column_types import UUID
from piccolo.columns.defaults.uuid import UUID7
from piccolo.table import Table
from piccolo.testing.test_case import AsyncTableTest
from tests.base import (
    engine_version_gte,
    is_running_postgres,
    python_version_gte,
)


class UUIDTable(Table):
    uuid = UUID()


class TestUUID(AsyncTableTest):
    tables = [UUIDTable]

    async def test_return_type(self):
        row = UUIDTable()
        await row.save()

        self.assertIsInstance(row.uuid, uuid.UUID)


class UUID7Table(Table):
    uuid_7 = UUID(default=UUID7())


@pytest.mark.skipif(
    not (
        python_version_gte(decimal.Decimal("3.14"))
        and is_running_postgres()
        and engine_version_gte(18)
    ),
    reason="Only Python >= 3.14 and Postgres >= 18 are supported.",
)
class TestUUID7(AsyncTableTest):
    tables = [UUID7Table]

    async def test_return_type(self):
        row = UUID7Table()
        await row.save()

        self.assertIsInstance(row.uuid_7, uuid.UUID)



import asyncio
import os
import tempfile
from typing import cast
from unittest import TestCase
from unittest.mock import call, patch

from piccolo.engine.postgres import PostgresEngine
from piccolo.engine.sqlite import SQLiteEngine
from tests.base import DBTestCase, engine_is, engines_only, sqlite_only
from tests.example_apps.music.tables import Manager


@engines_only("postgres", "cockroach")
class TestPool(DBTestCase):
    async def _create_pool(self) -> None:
        engine = cast(PostgresEngine, Manager._meta.db)

        await engine.start_connection_pool()
        assert engine.pool is not None

        await engine.close_connection_pool()
        assert engine.pool is None

    async def _make_query(self):
        await Manager._meta.db.start_connection_pool()

        await Manager(name="Bob").save().run()
        response = await Manager.select().run()
        self.assertIn("Bob", [i["name"] for i in response])

        await Manager._meta.db.close_connection_pool()

    async def _make_many_queries(self):
        await Manager._meta.db.start_connection_pool()

        await Manager(name="Bob").save().run()

        async def get_data():
            response = await Manager.select().run()
            if engine_is("cockroach"):
                self.assertEqual(
                    response, [{"id": response[0]["id"], "name": "Bob"}]
                )
            else:
                self.assertEqual(response, [{"id": 1, "name": "Bob"}])

        await asyncio.gather(*[get_data() for _ in range(500)])

        await Manager._meta.db.close_connection_pool()

    def test_creation(self):
        """
        Make sure a connection pool can be created.
        """
        asyncio.run(self._create_pool())

    def test_query(self):
        """
        Make several queries using a connection pool.
        """
        asyncio.run(self._make_query())

    def test_many_queries(self):
        """
        Make sure the connection pool is working correctly, and we don't
        exceed a connection limit - queries should queue, then succeed.
        """
        asyncio.run(self._make_many_queries())


@engines_only("postgres", "cockroach")
class TestPoolProxyMethods(DBTestCase):
    async def _create_pool(self) -> None:
        engine = cast(PostgresEngine, Manager._meta.db)

        # Deliberate typo ('nnn'):
        await engine.start_connnection_pool()
        assert engine.pool is not None

        # Deliberate typo ('nnn'):
        await engine.close_connnection_pool()
        assert engine.pool is None

    def test_proxy_methods(self):
        """
        There are some proxy methods, due to some old typos. Make sure they
        work, to ensure backwards compatibility.
        """
        asyncio.run(self._create_pool())


@sqlite_only
class TestConnectionPoolWarning(TestCase):
    async def _create_pool(self):
        sqlite_file = os.path.join(tempfile.gettempdir(), "engine.sqlite")
        engine = SQLiteEngine(path=sqlite_file)

        with patch("piccolo.engine.base.colored_warning") as colored_warning:
            await engine.start_connection_pool()
            await engine.close_connection_pool()

            self.assertEqual(
                colored_warning.call_args_list,
                [
                    call(
                        "Connection pooling is not supported for sqlite.",
                        stacklevel=3,
                    ),
                    call(
                        "Connection pooling is not supported for sqlite.",
                        stacklevel=3,
                    ),
                ],
            )

    def test_warnings(self):
        """
        Make sure that when trying to start and close a connection pool with
        SQLite, a warning is printed out, as connection pools aren't currently
        supported.
        """
        asyncio.run(self._create_pool())



from __future__ import annotations

import datetime
import decimal
import uuid
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, TypeVar, Union

from typing_extensions import TypeAlias

if TYPE_CHECKING:  # pragma: no cover
    from piccolo.columns.combination import And, Or, Where, WhereRaw  # noqa
    from piccolo.table import Table


Combinable = Union["Where", "WhereRaw", "And", "Or"]
CustomIterable = Iterable[Any]


TableInstance = TypeVar("TableInstance", bound="Table")
QueryResponseType = TypeVar("QueryResponseType", bound=Any)


# These are types we can reasonably expect to send to the database.
BasicTypes: TypeAlias = Union[
    bytes,
    datetime.date,
    datetime.datetime,
    datetime.time,
    datetime.timedelta,
    decimal.Decimal,
    dict,
    float,
    int,
    list,
    str,
    uuid.UUID,
]

###############################################################################
# For backwards compatibility:

from piccolo.columns.defaults.timestamp import DatetimeDefault  # noqa



import uuid
from collections.abc import Callable
from enum import Enum
from typing import Union

from piccolo.utils.uuid import uuid7

from .base import Default


class UUID4(Default):
    """
    This makes the default value for a
    :class:`UUID <piccolo.columns.column_types.UUID>` column a randomly
    generated UUID v4 value. The advantage over using :func:`uuid.uuid4` from
    the standard library, is the default is set on the column definition in the
    database too.
    """

    @property
    def postgres(self):
        """
        Historically we had to use `uuid_generate_v4()` from the `uuid-ossp`
        extension.

        Since Postgres 13 there is a built-in `gen_random_uuid` function which
        generates UUID v4 values.

        In Postgres 18, `uuidv4` was added, which is the same as
        `gen_random_uuid`, but more precisely named. We will move to this at
        some point in the future.

        """
        return "gen_random_uuid()"

    @property
    def cockroach(self):
        return self.postgres

    @property
    def sqlite(self):
        return None

    def python(self):
        return uuid.uuid4()


class UUID7(Default):
    """
    This makes the default value for a
    :class:`UUID <piccolo.columns.column_types.UUID>` column a randomly
    generated UUID v7 value. Postgres >= 18 and Python >= 3.14 only. UUID7 can
    be more efficiently indexed than UUID, and inserts can be faster.

    For this to work in older versions of Postgres, register an extension such
    as `pg_uuidv7 <https://github.com/fboulnois/pg_uuidv7>`_.

    """  # noqa: E501

    @property
    def postgres(self):
        """
        Supported in Postgres 18 and above.
        """
        return "uuidv7()"

    @property
    def cockroach(self):
        """
        CockroachDB doesn't currently have a uuidv7 function.
        """
        return None

    @property
    def sqlite(self):
        return None

    def python(self):
        return uuid7()


UUIDArg = Union[
    UUID4,
    UUID7,
    uuid.UUID,
    str,
    Enum,
    None,
    Callable[[], uuid.UUID],
]


__all__ = ["UUIDArg", "UUID4", "UUID7"]



from __future__ import annotations

import ast
import inspect
import itertools
import os
import pathlib
import traceback
from abc import abstractmethod
from collections.abc import Callable, Sequence
from dataclasses import dataclass, field
from graphlib import TopologicalSorter
from importlib import import_module
from types import ModuleType
from typing import Optional, Union, cast

import black

from piccolo.apps.migrations.auto.migration_manager import MigrationManager
from piccolo.engine.base import Engine
from piccolo.table import Table
from piccolo.utils.warnings import Level, colored_warning


class MigrationModule(ModuleType):
    ID: str
    VERSION: str
    DESCRIPTION: str

    @staticmethod
    @abstractmethod
    async def forwards() -> MigrationManager: ...


class PiccoloAppModule(ModuleType):
    APP_CONFIG: AppConfig


def get_package(name: str) -> str:
    """
    :param name:
        The __name__ variable from a Python file.

    """
    return ".".join(name.split(".")[:-1])


def table_finder(
    modules: Sequence[str],
    package: Optional[str] = None,
    include_tags: Optional[Sequence[str]] = None,
    exclude_tags: Optional[Sequence[str]] = None,
    exclude_imported: bool = False,
) -> list[type[Table]]:
    """
    Rather than explicitly importing and registering table classes with the
    ``AppConfig``, ``table_finder`` can be used instead. It imports any ``Table``
    subclasses in the given modules. Tags can be used to limit which ``Table``
    subclasses are imported.

    :param modules:
        The module paths to check for ``Table`` subclasses. For example,
        ``['blog.tables']``.
    :param package:
        This must be passed in if the modules are relative paths (e.g.
        if ``modules=['.tables']`` then ``package='blog'``).
    :param include_tags:
        If the ``Table`` subclass has one of these tags, it will be
        imported. The special tag ``'__all__'`` will import all ``Table``
        subclasses found.
    :param exclude_tags:
        If the ``Table`` subclass has any of these tags, it won't be
        imported. ``exclude_tags`` overrides ``include_tags``.
    :param exclude_imported:
        If ``True``, only ``Table`` subclasses defined within the module are
        used. Any ``Table`` subclasses imported by that module from other
        modules are ignored. For example:

        .. code-block:: python

            from piccolo.table import Table
            from piccolo.column import Varchar, ForeignKey
            from piccolo.apps.user.tables import BaseUser # excluded

            class Task(Table): # included
                title = Varchar()
                creator = ForeignKey(BaseUser)

    """  # noqa: E501
    if include_tags is None:
        include_tags = ["__all__"]
    if exclude_tags is None:
        exclude_tags = []
    if isinstance(modules, str):
        # Guard against the user just entering a string, for example
        # 'blog.tables', instead of ['blog.tables'].
        modules = [modules]

    table_subclasses: list[type[Table]] = []

    for module_path in modules:
        full_module_path = (
            ".".join([package, module_path.lstrip(".")])
            if package
            else module_path
        )

        try:
            module = import_module(
                module_path,
                package=package,
            )
        except ImportError as exception:
            print(f"Unable to import {full_module_path}")
            raise exception from exception

        object_names = [i for i in dir(module) if not i.startswith("_")]

        for object_name in object_names:
            _object = getattr(module, object_name)
            if (
                inspect.isclass(_object)
                and issubclass(_object, Table)
                and _object is not Table
            ):
                table: Table = _object  # type: ignore

                if exclude_imported and table.__module__ != full_module_path:
                    continue

                if exclude_tags and set(table._meta.tags).intersection(
                    set(exclude_tags)
                ):
                    continue
                elif "__all__" in include_tags:
                    table_subclasses.append(_object)
                elif set(table._meta.tags).intersection(set(include_tags)):
                    table_subclasses.append(_object)

    return table_subclasses


@dataclass
class Command:
    """
    :param callable:
        The function or method to be called.
    :param command_name:
        If not specified, the name of the ``callable`` is used.
    :param aliases:
        Alternative ways to refer to this command in the CLI.

    """

    callable: Callable
    command_name: Optional[str] = None
    aliases: list[str] = field(default_factory=list)


@dataclass
class AppConfig:
    """
    Each app needs an AppConfig, which is defined in piccolo_app.py.

    :param app_name:
        The name of the app, for example ``'article'``.
    :param migrations_folder_path:
        The path of the folder containing this app's migration files.
    :param table_classes:
        By registering table classes, Piccolo's auto migrations can detect
        changes to tables.
    :param migration_dependencies:
        A list of Piccolo apps whose migrations this app depends on. For
        example: ``['piccolo.apps.user.piccolo_conf']``. The migrations for
        those apps will be run before the migrations for this app.
    :param commands:
        A list of functions and coroutines, which are then registered with
        the Piccolo CLI. For example, with a Piccolo app called ``'article'``,
        and a command called ``new``, it can be called on the command line
        using ``piccolo article new``.

    """

    app_name: str
    migrations_folder_path: Union[str, pathlib.Path]
    table_classes: list[type[Table]] = field(default_factory=list)
    migration_dependencies: list[str] = field(default_factory=list)
    commands: list[Union[Callable, Command]] = field(default_factory=list)

    @property
    def resolved_migrations_folder_path(self) -> str:
        return (
            str(self.migrations_folder_path)
            if isinstance(self.migrations_folder_path, pathlib.Path)
            else self.migrations_folder_path
        )

    def __post_init__(self) -> None:
        self._migration_dependency_app_configs: Optional[list[AppConfig]] = (
            None
        )

    def register_table(self, table_class: type[Table]):
        self.table_classes.append(table_class)
        return table_class

    def get_commands(self) -> list[Command]:
        return [
            i if isinstance(i, Command) else Command(i) for i in self.commands
        ]

    @property
    def migration_dependency_app_configs(self) -> list[AppConfig]:
        """
        Get all of the ``AppConfig`` instances from this app's migration
        dependencies.
        """
        # We cache the value so it's more efficient, and also so we can set the
        # underlying value in unit tests for easier mocking.
        if self._migration_dependency_app_configs is None:
            modules: list[PiccoloAppModule] = [
                cast(PiccoloAppModule, import_module(module_path))
                for module_path in self.migration_dependencies
            ]
            self._migration_dependency_app_configs = [
                i.APP_CONFIG for i in modules
            ]

        return self._migration_dependency_app_configs

    def get_table_with_name(self, table_class_name: str) -> type[Table]:
        """
        Returns a ``Table`` subclass with the given name from this app, if it
        exists. Otherwise raises a ``ValueError``.
        """
        filtered = [
            table_class
            for table_class in self.table_classes
            if table_class.__name__ == table_class_name
        ]
        if not filtered:
            raise ValueError(
                f"No table with class name {table_class_name} exists."
            )
        return filtered[0]


class AppRegistry:
    """
    Records all of the Piccolo apps in your project. Kept in
    ``piccolo_conf.py``.

    :param apps:
        A list of paths to Piccolo apps, e.g. ``['blog.piccolo_app']``.

    """

    def __init__(self, apps: Optional[list[str]] = None):
        self.apps = apps or []
        self.app_configs: dict[str, AppConfig] = {}
        app_names = []

        for app in self.apps:
            try:
                app_conf_module = import_module(app)
                app_config: AppConfig = getattr(app_conf_module, "APP_CONFIG")
            except (ImportError, AttributeError) as e:
                if app.endswith(".piccolo_app"):
                    raise e from e
                app += ".piccolo_app"
                app_conf_module = import_module(app)
                app_config = getattr(app_conf_module, "APP_CONFIG")
                colored_warning(
                    f"App {app[:-12]} should end with `.piccolo_app`",
                    level=Level.medium,
                )

            self.app_configs[app_config.app_name] = app_config
            app_names.append(app_config.app_name)

        self._validate_app_names(app_names)

    @staticmethod
    def _validate_app_names(app_names: list[str]):
        """
        Raise a ValueError if an app_name is repeated.
        """
        app_names.sort()
        grouped = itertools.groupby(app_names)
        for key, value in grouped:
            count = len(list(value))
            if count > 1:
                raise ValueError(
                    f"There are {count} apps with the name `{key}`. This can "
                    "cause unexpected behavior. Make sure each app has a "
                    "unique name, and you haven't registered the same app "
                    "multiple times."
                )

    def get_app_config(self, app_name: str) -> Optional[AppConfig]:
        return self.app_configs.get(app_name)

    def get_table_classes(self, app_name: str) -> list[type[Table]]:
        """
        Returns each Table subclass defined in the given app if it exists.
        Otherwise raises a ValueError.

        :raises ValueError:
            If an AppConfig can't be found for the given app_name.

        """
        app_config = self.get_app_config(app_name=app_name)
        if not app_config:
            raise ValueError(f"Unrecognised app_name: {app_name}")
        return app_config.table_classes

    def get_table_with_name(
        self, app_name: str, table_class_name: str
    ) -> Optional[type[Table]]:
        """
        Returns a Table subclass registered with the given app if it exists.
        Otherwise raises a ValueError.
        """
        app_config = self.get_app_config(app_name=app_name)
        if app_config is None:
            raise ValueError(f"Can't find an app_config for {app_name}")
        else:
            return app_config.get_table_with_name(
                table_class_name=table_class_name
            )


class PiccoloConfModule(ModuleType):
    DB: Engine
    APP_REGISTRY: AppRegistry


DEFAULT_MODULE_NAME = "piccolo_conf"
ENVIRONMENT_VARIABLE = "PICCOLO_CONF"
ENGINE_VAR = "DB"


class Finder:
    """
    Contains useful methods for locating and loading apps within your project,
    and tables within apps.
    """

    def __init__(self, diagnose: bool = False):
        """
        :param diagnose:
            If True, when trying to import piccolo_conf, a traceback will be
            printed out if an error occurs.

        """
        self.diagnose = diagnose

    def _deduplicate(
        self, config_modules: list[PiccoloAppModule]
    ) -> list[PiccoloAppModule]:
        """
        Remove all duplicates - just leaving the first instance.
        """
        # Deduplicate, but preserve order - which is why set() isn't used.
        return list({c: None for c in config_modules}.keys())

    def _import_app_modules(
        self, config_module_paths: list[str]
    ) -> list[PiccoloAppModule]:
        """
        Import all piccolo_app.py modules within your apps, and all
        dependencies.
        """
        config_modules = []

        for config_module_path in config_module_paths:
            try:
                config_module = cast(
                    PiccoloAppModule, import_module(config_module_path)
                )
            except ImportError as e:
                raise Exception(
                    f"Unable to import {config_module_path}"
                ) from e
            app_config: AppConfig = getattr(config_module, "APP_CONFIG")
            dependency_config_modules = self._import_app_modules(
                app_config.migration_dependencies
            )
            config_modules.extend(dependency_config_modules + [config_module])

        return config_modules

    def get_piccolo_conf_module(
        self, module_name: Optional[str] = None
    ) -> Optional[PiccoloConfModule]:
        """
        Searches the path for a 'piccolo_conf.py' module to import. The
        location searched can be overriden by:

        * Explicitly passing a module name into this method.
        * Setting the PICCOLO_CONF environment variable.

        An example override is 'my_folder.piccolo_conf'.

        """
        env_module_name = os.environ.get(ENVIRONMENT_VARIABLE, None)

        if not module_name and env_module_name:
            module_name = env_module_name

        if not module_name:
            module_name = DEFAULT_MODULE_NAME

        try:
            module = cast(PiccoloConfModule, import_module(module_name))
        except ModuleNotFoundError as exc:
            if self.diagnose:
                colored_warning(
                    (
                        f"{module_name} either doesn't exist or the import "
                        "failed. Traceback:"
                    ),
                    level=Level.high,
                )
                print(traceback.format_exc())

            if str(exc) == "No module named 'asyncpg'":
                raise ModuleNotFoundError(
                    "PostgreSQL driver not found. "
                    "Try running `pip install 'piccolo[postgres]'`"
                ) from exc
            elif str(exc) == "No module named 'aiosqlite'":
                raise ModuleNotFoundError(
                    "SQLite driver not found. "
                    "Try running `pip install 'piccolo[sqlite]'`"
                ) from exc
            else:
                raise exc from exc
        else:
            return module

    def get_piccolo_conf_path(self) -> str:
        piccolo_conf_module = self.get_piccolo_conf_module()

        if piccolo_conf_module is None:
            raise ModuleNotFoundError("piccolo_conf.py not found.")

        module_file_path = piccolo_conf_module.__file__
        assert module_file_path

        return module_file_path

    def get_app_registry(self) -> AppRegistry:
        """
        Returns the ``AppRegistry`` instance within piccolo_conf.
        """
        piccolo_conf_module = self.get_piccolo_conf_module()
        return getattr(piccolo_conf_module, "APP_REGISTRY")

    def get_engine(
        self, module_name: Optional[str] = None
    ) -> Optional[Engine]:
        piccolo_conf = self.get_piccolo_conf_module(module_name=module_name)
        engine: Optional[Engine] = getattr(piccolo_conf, ENGINE_VAR, None)

        if not engine:
            colored_warning(
                f"{module_name} doesn't define a {ENGINE_VAR} variable.",
                level=Level.high,
            )
        elif not isinstance(engine, Engine):
            colored_warning(
                f"{module_name} contains a {ENGINE_VAR} variable of the "
                "wrong type - it should be an Engine subclass.",
                level=Level.high,
            )

        return engine

    def get_app_modules(self) -> list[PiccoloAppModule]:
        """
        Returns the ``piccolo_app.py`` modules for each registered Piccolo app
        in your project.
        """
        app_registry = self.get_app_registry()
        app_modules = self._import_app_modules(app_registry.apps)

        # Now deduplicate any dependencies
        app_modules = self._deduplicate(app_modules)

        return app_modules

    def get_app_names(
        self, sort_by_migration_dependencies: bool = True
    ) -> list[str]:
        """
        Return all of the app names.

        :param sort_by_migration_dependencies:
            If True, sorts the app names using the migration dependencies, so
            dependencies are before dependents in the list.

        """
        return [
            i.app_name
            for i in self.get_app_configs(
                sort_by_migration_dependencies=sort_by_migration_dependencies
            )
        ]

    def get_sorted_app_names(self) -> list[str]:
        """
        Just here for backwards compatibility - use ``get_app_names`` directly.
        """
        return self.get_app_names(sort_by_migration_dependencies=True)

    def sort_app_configs(
        self, app_configs: list[AppConfig]
    ) -> list[AppConfig]:
        app_config_map = {
            app_config.app_name: app_config for app_config in app_configs
        }

        sorted_app_names = TopologicalSorter(
            {
                app_config.app_name: [
                    i.app_name
                    for i in app_config.migration_dependency_app_configs
                ]
                for app_config in app_config_map.values()
            }
        ).static_order()

        return [app_config_map[i] for i in sorted_app_names]

    def get_app_configs(
        self, sort_by_migration_dependencies: bool = True
    ) -> list[AppConfig]:
        """
        Returns a list of ``AppConfig``, optionally sorted by migration
        dependencies.
        """
        app_configs = [i.APP_CONFIG for i in self.get_app_modules()]

        return (
            self.sort_app_configs(app_configs=app_configs)
            if sort_by_migration_dependencies
            else app_configs
        )

    def get_app_config(self, app_name: str) -> AppConfig:
        """
        Returns an ``AppConfig`` for the given app name.
        """
        for app_config in self.get_app_configs():
            if app_config.app_name == app_name:
                return app_config
        raise ValueError(f"No app found with name {app_name}")

    def get_table_with_name(
        self, app_name: str, table_class_name: str
    ) -> type[Table]:
        """
        Returns a ``Table`` class registered with the given app if it exists.
        Otherwise it raises an ``ValueError``.
        """
        app_config = self.get_app_config(app_name=app_name)
        return app_config.get_table_with_name(
            table_class_name=table_class_name
        )

    def get_table_classes(
        self,
        include_apps: Optional[list[str]] = None,
        exclude_apps: Optional[list[str]] = None,
    ) -> list[type[Table]]:
        """
        Returns all ``Table`` classes registered with the given apps. If
        ``include_apps`` is ``None``, then ``Table`` classes will be returned
        for all apps.
        """
        if include_apps and exclude_apps:
            raise ValueError("Only specify `include_apps` or `exclude_apps`.")

        if include_apps:
            app_names = include_apps
        else:
            app_names = self.get_app_names()
            if exclude_apps:
                app_names = [i for i in app_names if i not in exclude_apps]

        tables: list[type[Table]] = []

        for app_name in app_names:
            app_config = self.get_app_config(app_name=app_name)
            tables.extend(app_config.table_classes)

        return tables


###############################################################################


class PiccoloConfUpdater:

    def __init__(self, piccolo_conf_path: Optional[str] = None):
        """
        :param piccolo_conf_path:
            The path to the piccolo_conf.py (e.g. `./piccolo_conf.py`). If not
            passed in, we use our ``Finder`` class to get it.
        """
        self.piccolo_conf_path = (
            piccolo_conf_path or Finder().get_piccolo_conf_path()
        )

    def _modify_app_registry_src(self, src: str, app_module: str) -> str:
        """
        :param src:
            The contents of the ``piccolo_conf.py`` file.
        :param app_module:
            The app to add to the registry e.g. ``'music.piccolo_app'``.
        :returns:
            Updated Python source code string.

        """
        ast_root = ast.parse(src)

        parsing_successful = False

        for node in ast.walk(ast_root):
            if isinstance(node, ast.Call):
                if (
                    isinstance(node.func, ast.Name)
                    and node.func.id == "AppRegistry"
                ):
                    if len(node.keywords) > 0:
                        keyword = node.keywords[0]
                        if keyword.arg == "apps":
                            apps = keyword.value
                            if isinstance(apps, ast.List):
                                apps.elts.append(
                                    ast.Constant(app_module, kind="str")
                                )
                                parsing_successful = True
                                break

        if not parsing_successful:
            raise SyntaxError(
                "Unable to parse piccolo_conf.py - `AppRegistry(apps=...)` "
                "not found)."
            )

        new_contents = ast.unparse(ast_root)

        formatted_contents = black.format_str(
            new_contents, mode=black.FileMode(line_length=80)
        )

        return formatted_contents

    def register_app(self, app_module: str):
        """
        Adds the given app to the ``AppRegistry`` in ``piccolo_conf.py``.

        This is used by command line tools like:

        .. code-block:: bash

            piccolo app new my_app --register

        :param app_module:
            The module of the app, e.g. ``'music.piccolo_app'``.

        """
        with open(self.piccolo_conf_path) as f:
            piccolo_conf_src = f.read()

        new_contents = self._modify_app_registry_src(
            src=piccolo_conf_src, app_module=app_module
        )

        with open(self.piccolo_conf_path, "wt") as f:
            f.write(new_contents)



from piccolo.columns.column_types import UUID, Varchar
from piccolo.table import Table
from piccolo.testing.test_case import AsyncTableTest
from tests.example_apps.music.tables import Manager


class ManagerUUID(Table):
    id = UUID(primary_key=True)
    name = Varchar()


class TestInstanceEquality(AsyncTableTest):
    tables = [
        Manager,
        ManagerUUID,
    ]

    async def test_instance_equality(self) -> None:
        """
        Make sure instance equality works, for tables with a `Serial` primary
        key.
        """
        manager_1 = Manager(name="Guido")
        await manager_1.save()

        manager_2 = Manager(name="Graydon")
        await manager_2.save()

        self.assertEqual(manager_1, manager_1)
        self.assertNotEqual(manager_1, manager_2)

        # Try fetching the row from the database.
        manager_1_from_db = (
            await Manager.objects().where(Manager.id == manager_1.id).first()
        )
        self.assertEqual(manager_1, manager_1_from_db)
        self.assertNotEqual(manager_2, manager_1_from_db)

        # Try rows which haven't been saved yet.
        # They have no primary key value (because they use Serial columns
        # as the primary key), so they shouldn't be equal.
        self.assertNotEqual(Manager(), Manager())
        self.assertNotEqual(manager_1, Manager())

        # Make sure an object is equal to itself, even if not saved.
        manager_unsaved = Manager()
        self.assertEqual(manager_unsaved, manager_unsaved)

    async def test_instance_equality_uuid(self) -> None:
        """
        Make sure instance equality works, for tables with a `UUID` primary
        key.
        """
        manager_1 = ManagerUUID(name="Guido")
        await manager_1.save()

        manager_2 = ManagerUUID(name="Graydon")
        await manager_2.save()

        self.assertEqual(manager_1, manager_1)
        self.assertNotEqual(manager_1, manager_2)

        # Try fetching the row from the database.
        manager_1_from_db = (
            await ManagerUUID.objects()
            .where(ManagerUUID.id == manager_1.id)
            .first()
        )
        self.assertEqual(manager_1, manager_1_from_db)
        self.assertNotEqual(manager_2, manager_1_from_db)

        # Make sure an object is equal to itself, even if not saved.
        manager_unsaved = ManagerUUID()
        self.assertEqual(manager_unsaved, manager_unsaved)



from typing import Union

from typing_extensions import TypeAlias

from piccolo.columns.base import Column
from piccolo.querystring import QueryString

ArrayType: TypeAlias = Union[Column, QueryString, list[object]]
ArrayItemType: TypeAlias = Union[Column, QueryString, object]


class ArrayQueryString(QueryString):
    def __add__(self, array: ArrayType):
        """
        QueryString will use the ``+`` operator by default for addition, but
        for arrays we want to concatenate them instead.
        """
        return ArrayCat(array_1=self, array_2=array)

    def __radd__(self, array: ArrayType):
        return ArrayCat(array_1=array, array_2=self)


class ArrayCat(ArrayQueryString):
    def __init__(
        self,
        array_1: ArrayType,
        array_2: ArrayType,
    ):
        """
        Concatenate two arrays.

        :param array_1:
            These values will be at the start of the new array.
        :param array_2:
            These values will be at the end of the new array.

        """
        for value in (array_1, array_2):
            if isinstance(value, Column):
                engine_type = value._meta.engine_type
                if engine_type not in ("postgres", "cockroach"):
                    raise ValueError(
                        "Only Postgres and Cockroach support array "
                        "concatenation."
                    )

        super().__init__("array_cat({}, {})", array_1, array_2)


class ArrayAppend(ArrayQueryString):
    def __init__(self, array: ArrayType, value: ArrayItemType):
        """
        Append an element to the end of an array.

        :param column:
            Identifies the column.
        :param value:
            The value to append.

        """
        if isinstance(array, Column):
            engine_type = array._meta.engine_type
            if engine_type not in ("postgres", "cockroach"):
                raise ValueError(
                    "Only Postgres and Cockroach support array appending."
                )

        super().__init__("array_append({}, {})", array, value)


class ArrayPrepend(ArrayQueryString):
    def __init__(self, array: ArrayType, value: ArrayItemType):
        """
        Append an element to the beginning of an array.

        :param value:
            The value to prepend.
        :param column:
            Identifies the column.

        """
        if isinstance(array, Column):
            engine_type = array._meta.engine_type
            if engine_type not in ("postgres", "cockroach"):
                raise ValueError(
                    "Only Postgres and Cockroach support array prepending."
                )

        super().__init__("array_prepend({}, {})", value, array)


class ArrayReplace(ArrayQueryString):
    def __init__(
        self,
        array: ArrayType,
        old_value: ArrayItemType,
        new_value: ArrayItemType,
    ):
        """
        Replace each array element equal to the given value with a new value.

        :param column:
            Identifies the column.
        :param old_value:
            The old value to be replaced.
        :param new_value:
            The new value we are replacing with.

        """
        if isinstance(array, Column):
            engine_type = array._meta.engine_type
            if engine_type not in ("postgres", "cockroach"):
                raise ValueError(
                    "Only Postgres and Cockroach support array substitution."
                )

        super().__init__(
            "array_replace({}, {}, {})", array, old_value, new_value
        )


class ArrayRemove(ArrayQueryString):
    def __init__(self, array: ArrayType, value: ArrayItemType):
        """
        Remove all elements equal to the given value
        from the array (array must be one-dimensional).

        :param column:
            Identifies the column.
        :param value:
            The value to remove.

        """
        if isinstance(array, Column):
            engine_type = array._meta.engine_type
            if engine_type not in ("postgres", "cockroach"):
                raise ValueError(
                    "Only Postgres and Cockroach support array removing."
                )

        super().__init__("array_remove({}, {})", array, value)


__all__ = (
    "ArrayCat",
    "ArrayAppend",
    "ArrayPrepend",
    "ArrayReplace",
    "ArrayRemove",
)



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Union

from piccolo.custom_types import BasicTypes
from piccolo.querystring import QueryString

if TYPE_CHECKING:
    from piccolo.columns import Column


class Coalesce(QueryString):
    def __init__(
        self,
        *args: Union[Column, QueryString, BasicTypes],
        alias: Optional[str] = None,
    ):
        """
        Returns the first non-null value.

        Here's an example to try in the playground::

            >>> await Album.select(Album.release_date)
            [
                {'release_date': datetime.date(2021, 1, 1)},
                {'release_date': datetime.date(2025, 1, 1)},
                {'release_date': datetime.date(2022, 2, 2)},
                {'release_date': None}
            ]

        One of the values is null - we can specify a fallback value::

            >>> from piccolo.functions.conditional import Coalesce
            >>> await Album.select(
            ...     Coalesce(Album.release_date, datetime.date(2050, 1, 1))
            ... )
            [
                {'release_date': datetime.date(2021, 1, 1)},
                {'release_date': datetime.date(2025, 1, 1)},
                {'release_date': datetime.date(2022, 2, 2)},
                {'release_date': datetime.date(2050, 1, 1)}
            ]

        Or us this abbreviated syntax::

            >>> await Album.select(
            ...     Album.release_date | datetime.date(2050, 1, 1)
            ... )
            [
                {'release_date': datetime.date(2021, 1, 1)},
                {'release_date': datetime.date(2025, 1, 1)},
                {'release_date': datetime.date(2022, 2, 2)},
                {'release_date': datetime.date(2050, 1, 1)}
            ]

        """
        if len(args) < 2:
            raise ValueError("At least two values must be passed in.")

        #######################################################################
        # Preserve the original alias from the column.

        from piccolo.columns import Column

        first_arg = args[0]

        if isinstance(first_arg, Column):
            alias = (
                alias
                or first_arg._alias
                or first_arg._meta.get_default_alias()
            )
        elif isinstance(first_arg, QueryString):
            alias = alias or first_arg._alias

        #######################################################################

        placeholders = ", ".join("{}" for _ in args)

        super().__init__(f"COALESCE({placeholders})", *args, alias=alias)



from unittest import TestCase

from piccolo.apps.playground.commands.run import Genre, Manager


class TestTableStr(TestCase):
    def test_all_attributes(self):
        self.assertEqual(
            Manager._table_str(),
            (
                "class Manager(Table, tablename='manager'):\n"
                "    id = Serial(null=False, primary_key=True, unique=False, index=False, index_method=IndexMethod.btree, choices=None, db_column_name='id', secret=False)\n"  # noqa: E501
                "    name = Varchar(length=50, default='', null=False, primary_key=False, unique=False, index=False, index_method=IndexMethod.btree, choices=None, db_column_name=None, secret=False)\n"  # noqa: E501
            ),
        )

    def test_abbreviated(self):
        self.assertEqual(
            Manager._table_str(abbreviated=True),
            (
                "class Manager(Table):\n"
                "    id = Serial(primary_key=True)\n"
                "    name = Varchar(length=50)\n"
            ),
        )

    def test_m2m(self):
        """
        Make sure M2M relationships appear in the Table string.
        """

        self.assertEqual(
            Genre._table_str(abbreviated=True),
            (
                "class Genre(Table):\n"
                "    id = Serial(primary_key=True)\n"
                "    name = Varchar()\n"
                "    bands = M2M(GenreToBand)\n"
            ),
        )

    def test_print(self):
        """
        Make sure we can print it directly without any errors.
        """
        print(Manager)



from __future__ import annotations

import importlib
import os
import pathlib
import string
import sys
from typing import Any

import black
import jinja2

from piccolo.conf.apps import PiccoloConfUpdater

TEMPLATE_DIRECTORY = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "templates"
)

JINJA_ENV = jinja2.Environment(
    loader=jinja2.FileSystemLoader(searchpath=TEMPLATE_DIRECTORY),
)


def module_exists(module_name: str) -> bool:
    """
    Check whether a Python module already exists with this name - for
    example, a builtin module.
    """
    try:
        importlib.import_module(module_name)
    except ModuleNotFoundError:
        return False
    else:
        return True


APP_NAME_ALLOWED_CHARACTERS = [*string.ascii_lowercase, *string.digits, "_"]


def validate_app_name(app_name: str):
    """
    Make sure the app name is something which is a valid Python package name.

    :raises ValueError:
        If ``app_name`` isn't valid.

    """
    for char in app_name:
        if char.lower() not in APP_NAME_ALLOWED_CHARACTERS:
            raise ValueError(
                f"The app name contains a disallowed character: `{char}`. "
                "It must only include a-z, 0-9, and _ characters."
            )


def get_app_module(app_name: str, root: str) -> str:
    return ".".join([*pathlib.Path(root).parts, app_name, "piccolo_app"])


def new_app(app_name: str, root: str = ".", register: bool = False):
    print(f"Creating {app_name} app ...")

    try:
        validate_app_name(app_name=app_name)
    except ValueError as exception:
        sys.exit(str(exception))

    if module_exists(app_name):
        sys.exit(
            f"A module called {app_name} already exists - possibly a builtin "
            "Python module. Please choose a different name for your app."
        )

    app_root = os.path.join(root, app_name)

    if os.path.exists(app_root):
        sys.exit("Folder already exists - exiting.")

    os.makedirs(app_root)

    with open(os.path.join(app_root, "__init__.py"), "w"):
        pass

    templates: dict[str, Any] = {
        "piccolo_app.py": {"app_name": app_name},
        "tables.py": {},
    }

    for filename, context in templates.items():
        with open(os.path.join(app_root, filename), "w") as f:
            template = JINJA_ENV.get_template(f"{filename}.jinja")
            file_contents = template.render(**context)
            file_contents = black.format_str(
                file_contents, mode=black.FileMode(line_length=80)
            )
            f.write(file_contents)

    migrations_folder_path = os.path.join(app_root, "piccolo_migrations")
    os.mkdir(migrations_folder_path)

    with open(os.path.join(migrations_folder_path, "__init__.py"), "w"):
        pass

    if register:
        app_module = get_app_module(app_name=app_name, root=root)
        PiccoloConfUpdater().register_app(app_module=app_module)


def new(app_name: str, root: str = ".", register: bool = False):
    """
    Creates a new Piccolo app.

    :param app_name:
        The name of the new app.
    :param root:
        Where to create the app e.g. ./my/folder. By default it creates the
        app in the current directory.
    :param register:
        If True, the app is registered automatically in piccolo_conf.py.

    """
    new_app(app_name=app_name, root=root, register=register)



from unittest import IsolatedAsyncioTestCase

from piccolo.apps.migrations.tables import Migration
from piccolo.main import main


class TestMain(IsolatedAsyncioTestCase):

    async def asyncTearDown(self):
        await Migration.alter().drop_table(if_exists=True)

    def test_main(self):
        # Just make sure it runs without raising any errors.
        main()



from __future__ import annotations

from collections.abc import Sequence
from typing import Any, Optional, cast

from .postgres import Atomic, PostgresEngine, PostgresTransaction


class CockroachAtomic(Atomic):

    __slots__ = ("autocommit_before_ddl",)

    def __init__(
        self,
        engine: CockroachEngine,
        autocommit_before_ddl: Optional[bool] = False,
    ):
        """
        :param autocommit_before_ddl:
            Defaults to ``False`` to prevent automatic DDL commits
            in transactions (preventing rollbacks). Applies only to the current
            transaction and is automatically reverted when the transaction
            commits or is rolled back.

            Usage::

                # Defaults to ``False`` (``autocommit_before_ddl = off``)
                transaction = engine.atomic()
                transaction.add(Foo.create_table())

                # If we want to set ``autocommit_before_ddl = on``,
                # which is the default Cockroach session setting.
                transaction = engine.atomic(autocommit_before_ddl=True)
                transaction.add(Foo.create_table())

        """
        super().__init__(engine)
        self.autocommit_before_ddl = autocommit_before_ddl

    async def setup_transaction(self, transaction: PostgresTransaction):
        if self.autocommit_before_ddl is not None:
            transaction = cast(CockroachTransaction, transaction)
            await transaction.autocommit_before_ddl(
                enabled=self.autocommit_before_ddl
            )


class CockroachTransaction(PostgresTransaction):

    async def autocommit_before_ddl(self, enabled: bool = True):
        value = "on" if enabled else "off"
        await self.connection.execute(
            f"SET LOCAL autocommit_before_ddl = {value}"
        )


class CockroachEngine(PostgresEngine):
    """
    An extension of
    :class:`PostgresEngine <piccolo.engine.postgres.PostgresEngine>`.
    """

    def __init__(
        self,
        config: dict[str, Any],
        extensions: Sequence[str] = (),
        log_queries: bool = False,
        log_responses: bool = False,
        extra_nodes: Optional[dict[str, CockroachEngine]] = None,
    ) -> None:
        super().__init__(
            config=config,
            extensions=extensions,
            log_queries=log_queries,
            log_responses=log_responses,
            extra_nodes=extra_nodes,
        )
        self.engine_type = "cockroach"
        self.min_version_number = 0

    def atomic(
        self,
        autocommit_before_ddl: Optional[bool] = False,
    ) -> CockroachAtomic:
        return CockroachAtomic(
            engine=self,
            autocommit_before_ddl=autocommit_before_ddl,
        )

    def transaction(
        self,
        allow_nested: bool = True,
    ) -> CockroachTransaction:
        return CockroachTransaction(
            engine=self,
            allow_nested=allow_nested,
        )



from __future__ import annotations

import datetime
import decimal
import json
from collections.abc import Callable
from decimal import Decimal
from typing import Any, Optional, Union, cast
from uuid import UUID

from piccolo.columns import JSON, JSONB, Array, Column, ForeignKey
from piccolo.custom_types import TableInstance
from piccolo.testing.random_builder import RandomBuilder
from piccolo.utils.sync import run_sync


class ModelBuilder:
    __DEFAULT_MAPPER: dict[type, Callable] = {
        bool: RandomBuilder.next_bool,
        bytes: RandomBuilder.next_bytes,
        datetime.date: RandomBuilder.next_date,
        datetime.datetime: RandomBuilder.next_datetime,
        float: RandomBuilder.next_float,
        decimal.Decimal: RandomBuilder.next_decimal,
        int: RandomBuilder.next_int,
        str: RandomBuilder.next_str,
        datetime.time: RandomBuilder.next_time,
        datetime.timedelta: RandomBuilder.next_timedelta,
        UUID: RandomBuilder.next_uuid,
    }

    @classmethod
    async def build(
        cls,
        table_class: type[TableInstance],
        defaults: Optional[dict[Union[Column, str], Any]] = None,
        persist: bool = True,
        minimal: bool = False,
    ) -> TableInstance:
        """
        Build a ``Table`` instance with random data and save async.
        If the ``Table`` has any foreign keys, then the related rows are also
        created automatically.

        :param table_class:
            Table class to randomize.
        :param defaults:
            Any values specified here will be used instead of random values.
        :param persist:
            Whether to save the new instance in the database.
        :param minimal:
            If ``True`` then any columns with ``null=True`` are assigned
            a value of ``None``.

        Examples::

            # Create a new instance with all random values:
            manager = await ModelBuilder.build(Manager)

            # Create a new instance, with certain defaults:
            manager = await ModelBuilder.build(
                Manager,
                {Manager.name: 'Guido'}
            )

            # Create a new instance, but don't save it in the database:
            manager = await ModelBuilder.build(Manager, persist=False)

            # Create a new instance, with all null values set to None:
            manager = await ModelBuilder.build(Manager, minimal=True)

            # We can pass other table instances in as default values:
            band = await ModelBuilder.build(Band, {Band.manager: manager})

        """
        return await cls._build(
            table_class=table_class,
            defaults=defaults,
            persist=persist,
            minimal=minimal,
        )

    @classmethod
    def build_sync(
        cls,
        table_class: type[TableInstance],
        defaults: Optional[dict[Union[Column, str], Any]] = None,
        persist: bool = True,
        minimal: bool = False,
    ) -> TableInstance:
        """
        A sync wrapper around :meth:`build`.
        """
        return run_sync(
            cls.build(
                table_class=table_class,
                defaults=defaults,
                persist=persist,
                minimal=minimal,
            )
        )

    @classmethod
    async def _build(
        cls,
        table_class: type[TableInstance],
        defaults: Optional[dict[Union[Column, str], Any]] = None,
        minimal: bool = False,
        persist: bool = True,
    ) -> TableInstance:
        model = table_class(_ignore_missing=True)
        defaults = {} if not defaults else defaults

        for column, value in defaults.items():
            if isinstance(column, str):
                column = model._meta.get_column_by_name(column)

            setattr(model, column._meta.name, value)

        for column in model._meta.columns:
            if column._meta.null and minimal:
                continue

            if column._meta.name in defaults:
                continue  # Column value exists

            if isinstance(column, ForeignKey) and persist:
                # Check for recursion
                if column._foreign_key_meta.references is table_class:
                    if column._meta.null is True:
                        # We can avoid this problem entirely by setting it to
                        # None.
                        random_value = None
                    else:
                        # There's no way to avoid recursion in the situation.
                        raise ValueError("Recursive foreign key detected")
                else:
                    reference_model = await cls._build(
                        column._foreign_key_meta.resolved_references,
                        persist=True,
                    )
                    random_value = getattr(
                        reference_model,
                        reference_model._meta.primary_key._meta.name,
                    )
            else:
                random_value = cls._randomize_attribute(column)

            setattr(model, column._meta.name, random_value)

        if persist:
            await model.save().run()

        return model

    @classmethod
    def _randomize_attribute(cls, column: Column) -> Any:
        """
        Generate a random value for a column and apply formatting.

        :param column:
            Column class to randomize.

        """
        random_value: Any
        if column.value_type == Decimal:
            precision, scale = column._meta.params["digits"] or (4, 2)
            random_value = RandomBuilder.next_decimal(
                precision=precision, scale=scale
            )
        elif column.value_type == datetime.datetime:
            tz_aware = getattr(column, "tz_aware", False)
            random_value = RandomBuilder.next_datetime(tz_aware=tz_aware)
        elif column.value_type == list:
            length = RandomBuilder.next_int(maximum=10)
            if column._meta.choices:
                random_value = [
                    RandomBuilder.next_enum(column._meta.choices)
                    for _ in range(length)
                ]
            else:
                base_type = cast(Array, column).base_column.value_type
                random_value = [
                    cls.__DEFAULT_MAPPER[base_type]() for _ in range(length)
                ]
        elif column._meta.choices:
            random_value = RandomBuilder.next_enum(column._meta.choices)
        else:
            random_value = cls.__DEFAULT_MAPPER[column.value_type]()

        if "length" in column._meta.params and isinstance(random_value, str):
            return random_value[: column._meta.params["length"]]
        elif isinstance(column, (JSON, JSONB)):
            return json.dumps({"value": random_value})

        return random_value



"""
Notes for devs
==============

Descriptors
-----------

Each column type implements the descriptor protocol (the ``__get__`` and
``__set__`` methods).

This is to signal to MyPy that the following is allowed:

.. code-block:: python

    class Band(Table):
        name = Varchar()

    band = Band()
    band.name = 'Pythonistas'  # Without descriptors, this would be an error

In the above example, descriptors allow us to tell MyPy that ``name`` is a
``Varchar`` when accessed on a class, but is a ``str`` when accessed on a class
instance.

"""

from __future__ import annotations

import copy
import decimal
import inspect
import uuid
from collections.abc import Callable
from dataclasses import dataclass
from datetime import date, datetime, time, timedelta
from enum import Enum
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Literal,
    Optional,
    Union,
    cast,
    overload,
)

from typing_extensions import Unpack

from piccolo.columns.base import (
    Column,
    ColumnKwargs,
    ForeignKeyMeta,
    OnDelete,
    OnUpdate,
    ReferencedTable,
)
from piccolo.columns.combination import Where
from piccolo.columns.defaults.date import DateArg, DateCustom, DateNow
from piccolo.columns.defaults.interval import IntervalArg, IntervalCustom
from piccolo.columns.defaults.time import TimeArg, TimeCustom, TimeNow
from piccolo.columns.defaults.timestamp import (
    TimestampArg,
    TimestampCustom,
    TimestampNow,
)
from piccolo.columns.defaults.timestamptz import (
    TimestamptzArg,
    TimestamptzCustom,
    TimestamptzNow,
)
from piccolo.columns.defaults.uuid import UUID4, UUIDArg
from piccolo.columns.operators.comparison import (
    ArrayAll,
    ArrayAny,
    ArrayNotAny,
)
from piccolo.columns.operators.string import Concat
from piccolo.columns.reference import LazyTableReference
from piccolo.querystring import QueryString
from piccolo.utils.encoding import dump_json
from piccolo.utils.warnings import colored_warning

if TYPE_CHECKING:  # pragma: no cover
    from piccolo.columns.base import ColumnMeta
    from piccolo.query.functions.array import ArrayItemType, ArrayType
    from piccolo.query.operators.json import (
        GetChildElement,
        GetElementFromPath,
    )
    from piccolo.table import Table


###############################################################################


class ConcatDelegate:
    """
    Used in update queries to concatenate two strings - for example::

        await Band.update({Band.name: Band.name + 'abc'})

    """

    def get_querystring(
        self,
        column: Column,
        value: Union[str, Column, QueryString],
        reverse: bool = False,
    ) -> QueryString:
        """
        :param reverse:
            By default the value is appended to the column's value. If
            ``reverse=True`` then the value is prepended to the column's
            value instead.

        """
        if isinstance(value, Column):
            if len(column._meta.call_chain) > 0:
                raise ValueError(
                    "Adding values across joins isn't currently supported."
                )
        elif isinstance(value, str):
            value = QueryString("CAST({} AS TEXT)", value)
        elif not isinstance(value, QueryString):
            raise ValueError(
                "Only str, Column and QueryString values can be added."
            )

        args = [value, column] if reverse else [column, value]

        # We use the concat operator instead of the concat function, because
        # this is what we historically used, and they treat null values
        # differently.
        return QueryString(
            Concat.template.format(value_1="{}", value_2="{}"), *args
        )


class MathDelegate:
    """
    Used in update queries to perform math operations on columns, for example::

        await Band.update({Band.popularity: Band.popularity + 100})

    """

    def get_querystring(
        self,
        column_name: str,
        operator: Literal["+", "-", "/", "*"],
        value: Union[int, float, Integer],
        reverse: bool = False,
    ) -> QueryString:
        if isinstance(value, Integer):
            column: Integer = value
            if len(column._meta.call_chain) > 0:
                raise ValueError(
                    "Adding values across joins isn't currently supported."
                )
            other_column_name = value._meta.db_column_name
            return QueryString(f"{column_name} {operator} {other_column_name}")
        elif isinstance(value, (int, float)):
            if reverse:
                return QueryString(f"{{}} {operator} {column_name}", value)
            else:
                return QueryString(f"{column_name} {operator} {{}}", value)
        else:
            raise ValueError(
                "Only integers, floats, and other Integer columns can be "
                "added."
            )


class TimedeltaDelegate:
    """
    Used in update queries to add a timedelta to these columns:

    * ``Timestamp``
    * ``Timestamptz``
    * ``Date``
    * ``Interval``

    Example::

        class Concert(Table):
            starts = Timestamp()

        # Lets us increase all of the matching values by 1 day:
        >>> await Concert.update({
        ...     Concert.starts: Concert.starts + datetime.timedelta(days=1)
        ... })

    """

    # Maps the attribute name in Python's timedelta to what it's called in
    # Postgres.
    postgres_attr_map: dict[str, str] = {
        "days": "DAYS",
        "seconds": "SECONDS",
        "microseconds": "MICROSECONDS",
    }

    def get_postgres_interval_string(self, interval: timedelta) -> str:
        """
        :returns:
            A string like::

                "'1 DAYS 5 SECONDS 1000 MICROSECONDS'"

        """
        output = []
        for timedelta_key, postgres_name in self.postgres_attr_map.items():
            timestamp_value = getattr(interval, timedelta_key)
            if timestamp_value:
                output.append(f"{timestamp_value} {postgres_name}")

        output_string = " ".join(output)
        return f"'{output_string}'"

    def get_sqlite_interval_string(self, interval: timedelta) -> str:
        """
        :returns:
            A string like::

                "'+1 DAYS', '+5.001 SECONDS'"

        """
        output = []

        data = {
            "DAYS": interval.days,
            "SECONDS": interval.seconds + (interval.microseconds / 10**6),
        }

        for key, value in data.items():
            if value:
                operator = "+" if value >= 0 else ""
                output.append(f"'{operator}{value} {key}'")

        output_string = ", ".join(output)
        return output_string

    def get_querystring(
        self,
        column: Column,
        operator: Literal["+", "-"],
        value: timedelta,
        engine_type: str,
    ) -> QueryString:
        column_name = column._meta.name

        if not isinstance(value, timedelta):
            raise ValueError("Only timedelta values can be added.")

        if engine_type in ("postgres", "cockroach"):
            value_string = self.get_postgres_interval_string(interval=value)
            return QueryString(
                f'"{column_name}" {operator} INTERVAL {value_string}',
            )
        elif engine_type == "sqlite":
            if isinstance(column, Interval):
                # SQLite doesn't have a proper Interval type. Instead we store
                # the number of seconds.
                return QueryString(
                    f'CAST("{column_name}" AS REAL) {operator} {value.total_seconds()}'  # noqa: E501
                )
            elif isinstance(column, (Timestamp, Timestamptz)):
                if (
                    round(value.microseconds / 1000) * 1000
                    != value.microseconds
                ):
                    raise ValueError(
                        "timedeltas with such high precision won't save "
                        "accurately - the max resolution is 1 millisecond."
                    )
                strftime_format = "%Y-%m-%d %H:%M:%f"
            elif isinstance(column, Date):
                strftime_format = "%Y-%m-%d"
            else:
                raise ValueError(
                    f"{column.__class__.__name__} doesn't support timedelta "
                    "addition currently."
                )

            if operator == "-":
                value = value * -1

            value_string = self.get_sqlite_interval_string(interval=value)

            # We use `strftime` instead of `datetime`, because `datetime`
            # doesn't return microseconds.
            return QueryString(
                f"strftime('{strftime_format}', \"{column_name}\", {value_string})"  # noqa: E501
            )
        else:
            raise ValueError("Unrecognised engine")


###############################################################################


class Varchar(Column):
    """
    Used for storing text when you want to enforce character length limits.
    Uses the ``str`` type for values.

    **Example**

    .. code-block:: python

        class Band(Table):
            name = Varchar(length=100)

        # Create
        >>> await Band(name='Pythonistas').save()

        # Query
        >>> await Band.select(Band.name)
        {'name': 'Pythonistas'}

    :param length:
        The maximum number of characters allowed.

    """

    value_type = str
    concat_delegate: ConcatDelegate = ConcatDelegate()

    def __init__(
        self,
        length: Optional[int] = 255,
        default: Union[str, Enum, Callable[[], str], None] = "",
        **kwargs: Unpack[ColumnKwargs],
    ) -> None:
        self._validate_default(default, (str, None))

        self.length = length
        self.default = default
        super().__init__(length=length, default=default, **kwargs)

    @property
    def column_type(self):
        return f"VARCHAR({self.length})" if self.length else "VARCHAR"

    ###########################################################################
    # For update queries

    def __add__(self, value: Union[str, Varchar, Text]) -> QueryString:
        return self.concat_delegate.get_querystring(
            column=self,
            value=value,
        )

    def __radd__(self, value: Union[str, Varchar, Text]) -> QueryString:
        return self.concat_delegate.get_querystring(
            column=self,
            value=value,
            reverse=True,
        )

    ###########################################################################
    # Descriptors

    @overload
    def __get__(self, obj: Table, objtype=None) -> str: ...

    @overload
    def __get__(self, obj: None, objtype=None) -> Varchar: ...

    def __get__(self, obj, objtype=None):
        return obj.__dict__[self._meta.name] if obj else self

    def __set__(self, obj, value: Union[str, None]):
        obj.__dict__[self._meta.name] = value


class Email(Varchar):
    """
    Used for storing email addresses. It's identical to :class:`Varchar`,
    except when using :func:`create_pydantic_model <piccolo.utils.pydantic.create_pydantic_model>` -
    we add email validation to the Pydantic model. This means that :ref:`PiccoloAdmin`
    also validates email addresses.
    """  # noqa: E501

    pass


class Secret(Varchar):
    """
    This is just an alias to ``Varchar(secret=True)``. It's here for backwards
    compatibility.
    """

    def __init__(self, *args, **kwargs):
        kwargs["secret"] = True
        super().__init__(*args, **kwargs)

    ###########################################################################
    # Descriptors

    @overload
    def __get__(self, obj: Table, objtype=None) -> str: ...

    @overload
    def __get__(self, obj: None, objtype=None) -> Secret: ...

    def __get__(self, obj, objtype=None):
        return obj.__dict__[self._meta.name] if obj else self

    def __set__(self, obj, value: Union[str, None]):
        obj.__dict__[self._meta.name] = value


class Text(Column):
    """
    Use when you want to store large strings, and don't want to limit the
    string size. Uses the ``str`` type for values.

    **Example**

    .. code-block:: python

        class Band(Table):
            name = Text()

        # Create
        >>> await Band(name='Pythonistas').save()

        # Query
        >>> await Band.select(Band.name)
        {'name': 'Pythonistas'}

    """

    value_type = str
    concat_delegate: ConcatDelegate = ConcatDelegate()

    def __init__(
        self,
        default: Union[str, Enum, None, Callable[[], str]] = "",
        **kwargs: Unpack[ColumnKwargs],
    ) -> None:
        self._validate_default(default, (str, None))
        self.default = default
        super().__init__(default=default, **kwargs)

    ###########################################################################
    # For update queries

    def __add__(self, value: Union[str, Varchar, Text]) -> QueryString:
        return self.concat_delegate.get_querystring(
            column=self,
            value=value,
        )

    def __radd__(self, value: Union[str, Varchar, Text]) -> QueryString:
        return self.concat_delegate.get_querystring(
            column=self,
            value=value,
            reverse=True,
        )

    ###########################################################################
    # Descriptors

    @overload
    def __get__(self, obj: Table, objtype=None) -> str: ...

    @overload
    def __get__(self, obj: None, objtype=None) -> Text: ...

    def __get__(self, obj, objtype=None):
        return obj.__dict__[self._meta.name] if obj else self

    def __set__(self, obj, value: Union[str, None]):
        obj.__dict__[self._meta.name] = value


class UUID(Column):
    """
    Used for storing UUIDs - in Postgres a UUID column type is used, and in
    SQLite it's just a Varchar. Uses the ``uuid.UUID`` type for values.

    **Example**

    .. code-block:: python

        import uuid

        class Band(Table):
            uuid = UUID()

        # Create
        >>> await DiscountCode(code=uuid.uuid4()).save()

        # Query
        >>> await DiscountCode.select(DiscountCode.code)
        {'code': UUID('09c4c17d-af68-4ce7-9955-73dcd892e462')}

    """

    value_type = uuid.UUID

    def __init__(
        self,
        default: UUIDArg = UUID4(),
        **kwargs: Unpack[ColumnKwargs],
    ) -> None:
        if default is UUID4:
            # In case the class is passed in, instead of an instance.
            default = UUID4()

        self._validate_default(default, UUIDArg.__args__)  # type: ignore

        if default == uuid.uuid4:
            default = UUID4()

        if isinstance(default, str):
            try:
                default = uuid.UUID(default)
            except ValueError as e:
                raise ValueError(
                    "The default is a string, but not a valid uuid."
                ) from e

        self.default = default
        super().__init__(default=default, **kwargs)

    ###########################################################################
    # Descriptors

    @overload
    def __get__(self, obj: Table, objtype=None) -> uuid.UUID: ...

    @overload
    def __get__(self, obj: None, objtype=None) -> UUID: ...

    def __get__(self, obj, objtype=None):
        return obj.__dict__[self._meta.name] if obj else self

    def __set__(self, obj, value: Union[uuid.UUID, None]):
        obj.__dict__[self._meta.name] = value


class Integer(Column):
    """
    Used for storing whole numbers. Uses the ``int`` type for values.

    **Example**

    .. code-block:: python

        class Band(Table):
            popularity = Integer()

        # Create
        >>> await Band(popularity=1000).save()

        # Query
        >>> await Band.select(Band.popularity)
        {'popularity': 1000}

    """

    math_delegate = MathDelegate()

    def __init__(
        self,
        default: Union[int, Enum, Callable[[], int], None] = 0,
        **kwargs: Unpack[ColumnKwargs],
    ) -> None:
        self._validate_default(default, (int, None))
        self.default = default
        super().__init__(default=default, **kwargs)

    ###########################################################################
    # For update queries

    def __add__(self, value: Union[int, float, Integer]) -> QueryString:
        return self.math_delegate.get_querystring(
            column_name=self._meta.db_column_name, operator="+", value=value
        )

    def __radd__(self, value: Union[int, float, Integer]) -> QueryString:
        return self.math_delegate.get_querystring(
            column_name=self._meta.db_column_name,
            operator="+",
            value=value,
            reverse=True,
        )

    def __sub__(self, value: Union[int, float, Integer]) -> QueryString:
        return self.math_delegate.get_querystring(
            column_name=self._meta.db_column_name, operator="-", value=value
        )

    def __rsub__(self, value: Union[int, float, Integer]) -> QueryString:
        return self.math_delegate.get_querystring(
            column_name=self._meta.db_column_name,
            operator="-",
            value=value,
            reverse=True,
        )

    def __mul__(self, value: Union[int, float, Integer]) -> QueryString:
        return self.math_delegate.get_querystring(
            column_name=self._meta.db_column_name, operator="*", value=value
        )

    def __rmul__(self, value: Union[int, float, Integer]) -> QueryString:
        return self.math_delegate.get_querystring(
            column_name=self._meta.db_column_name,
            operator="*",
            value=value,
            reverse=True,
        )

    def __truediv__(self, value: Union[int, float, Integer]) -> QueryString:
        return self.math_delegate.get_querystring(
            column_name=self._meta.db_column_name, operator="/", value=value
        )

    def __rtruediv__(self, value: Union[int, float, Integer]) -> QueryString:
        return self.math_delegate.get_querystring(
            column_name=self._meta.db_column_name,
            operator="/",
            value=value,
            reverse=True,
        )

    def __floordiv__(self, value: Union[int, float, Integer]) -> QueryString:
        return self.math_delegate.get_querystring(
            column_name=self._meta.db_column_name, operator="/", value=value
        )

    def __rfloordiv__(self, value: Union[int, float, Integer]) -> QueryString:
        return self.math_delegate.get_querystring(
            column_name=self._meta.db_column_name,
            operator="/",
            value=value,
            reverse=True,
        )

    ###########################################################################
    # Descriptors

    @overload
    def __get__(self, obj: Table, objtype=None) -> int: ...

    @overload
    def __get__(self, obj: None, objtype=None) -> Integer: ...

    def __get__(self, obj, objtype=None):
        return obj.__dict__[self._meta.name] if obj else self

    def __set__(self, obj, value: Union[int, None]):
        obj.__dict__[self._meta.name] = value


###############################################################################
# BigInt and SmallInt only exist on Postgres. SQLite treats them the same as
# Integer columns.


class BigInt(Integer):
    """
    In Postgres, this column supports large integers. In SQLite, it's an alias
    to an Integer column, which already supports large integers. Uses the
    ``int`` type for values.

    **Example**

    .. code-block:: python

        class Band(Table):
            value = BigInt()

        # Create
        >>> await Band(popularity=1000000).save()

        # Query
        >>> await Band.select(Band.popularity)
        {'popularity': 1000000}

    """

    def _get_column_type(self, engine_type: str):
        if engine_type == "postgres":
            return "BIGINT"
        elif engine_type == "cockroach":
            return "BIGINT"
        elif engine_type == "sqlite":
            return "INTEGER"
        raise Exception("Unrecognized engine type")

    @property
    def column_type(self):
        return self._get_column_type(engine_type=self._meta.engine_type)

    ###########################################################################
    # Descriptors

    @overload
    def __get__(self, obj: Table, objtype=None) -> int: ...

    @overload
    def __get__(self, obj: None, objtype=None) -> BigInt: ...

    def __get__(self, obj, objtype=None):
        return obj.__dict__[self._meta.name] if obj else self

    def __set__(self, obj, value: Union[int, None]):
        obj.__dict__[self._meta.name] = value


class SmallInt(Integer):
    """
    In Postgres, this column supports small integers. In SQLite, it's an alias
    to an Integer column. Uses the ``int`` type for values.

    **Example**

    .. code-block:: python

        class Band(Table):
            value = SmallInt()

        # Create
        >>> await Band(popularity=1000).save()

        # Query
        >>> await Band.select(Band.popularity)
        {'popularity': 1000}

    """

    @property
    def column_type(self):
        engine_type = self._meta.engine_type
        if engine_type == "postgres":
            return "SMALLINT"
        elif engine_type == "cockroach":
            return "SMALLINT"
        elif engine_type == "sqlite":
            return "INTEGER"
        raise Exception("Unrecognized engine type")

    ###########################################################################
    # Descriptors

    @overload
    def __get__(self, obj: Table, objtype=None) -> int: ...

    @overload
    def __get__(self, obj: None, objtype=None) -> SmallInt: ...

    def __get__(self, obj, objtype=None):
        return obj.__dict__[self._meta.name] if obj else self

    def __set__(self, obj, value: Union[int, None]):
        obj.__dict__[self._meta.name] = value


###############################################################################


DEFAULT = QueryString("DEFAULT")
NULL = QueryString("null")


class Serial(Column):
    """
    An alias to an autoincrementing integer column in Postgres.
    """

    @property
    def column_type(self):
        engine_type = self._meta.engine_type
        if engine_type == "postgres":
            return "SERIAL"
        elif engine_type == "cockroach":
            return "INTEGER"
        elif engine_type == "sqlite":
            return "INTEGER"
        raise Exception("Unrecognized engine type")

    def default(self) -> QueryString:
        engine_type = self._meta.engine_type

        if engine_type == "postgres":
            return DEFAULT
        elif engine_type == "cockroach":
            return QueryString("unique_rowid()")
        elif engine_type == "sqlite":
            return NULL
        raise Exception("Unrecognized engine type")

    ###########################################################################
    # Descriptors

    @overload
    def __get__(self, obj: Table, objtype=None) -> int: ...

    @overload
    def __get__(self, obj: None, objtype=None) -> Serial: ...

    def __get__(self, obj, objtype=None):
        return obj.__dict__[self._meta.name] if obj else self

    def __set__(self, obj, value: Union[int, None]):
        obj.__dict__[self._meta.name] = value


class BigSerial(Serial):
    """
    An alias to a large autoincrementing integer column in Postgres.
    """

    @property
    def column_type(self):
        engine_type = self._meta.engine_type
        if engine_type == "postgres":
            return "BIGSERIAL"
        elif engine_type == "cockroach":
            return "BIGINT"
        elif engine_type == "sqlite":
            return "INTEGER"
        raise Exception("Unrecognized engine type")

    ###########################################################################
    # Descriptors

    @overload
    def __get__(self, obj: Table, objtype=None) -> int: ...

    @overload
    def __get__(self, obj: None, objtype=None) -> BigSerial: ...

    def __get__(self, obj, objtype=None):
        return obj.__dict__[self._meta.name] if obj else self

    def __set__(self, obj, value: Union[int, None]):
        obj.__dict__[self._meta.name] = value


class PrimaryKey(Serial):
    def __init__(
        self,
        **kwargs: Unpack[ColumnKwargs],
    ) -> None:
        # Set the index to False, as a database should automatically create
        # an index for a PrimaryKey column.
        kwargs.update({"primary_key": True, "index": False})

        colored_warning(
            "`PrimaryKey` is deprecated and will be removed in future "
            "versions. Use `UUID(primary_key=True)` or "
            "`Serial(primary_key=True)` instead. If no primary key column is "
            "specified, Piccolo will automatically add one for you called "
            "`id`.",
            category=DeprecationWarning,
        )

        super().__init__(**kwargs)

    ###########################################################################
    # Descriptors

    @overload
    def __get__(self, obj: Table, objtype=None) -> int: ...

    @overload
    def __get__(self, obj: None, objtype=None) -> PrimaryKey: ...

    def __get__(self, obj, objtype=None):
        return obj.__dict__[self._meta.name] if obj else self

    def __set__(self, obj, value: Union[int, None]):
        obj.__dict__[self._meta.name] = value


###############################################################################


class Timestamp(Column):
    """
    Used for storing timezone naive datetimes. Uses the ``datetime`` type for
    values.

    **Example**

    .. code-block:: python

        import datetime

        class Concert(Table):
            starts = Timestamp()

        # Create
        >>> await Concert(
        ...    starts=datetime.datetime(year=2050, month=1, day=1)
        ... ).save()

        # Query
        >>> await Concert.select(Concert.starts)
        {'starts': datetime.datetime(2050, 1, 1, 0, 0)}

    .. note::
        We recommend using
        :class:`Timestamptz <piccolo.columns.column_types.Timestamptz>`
        columns if possible - having timezone information makes the timestamp
        unambiguous.

    """

    value_type = datetime
    timedelta_delegate = TimedeltaDelegate()

    def __init__(
        self,
        default: TimestampArg = TimestampNow(),
        **kwargs: Unpack[ColumnKwargs],
    ) -> None:
        self._validate_default(default, TimestampArg.__args__)  # type: ignore

        if isinstance(default, datetime):
            if default.tzinfo is not None:
                raise ValueError(
                    "Timestamp only stores timezone naive datetime objects - "
                    "use Timestamptz instead."
                )
            default = TimestampCustom.from_datetime(default)

        if default == datetime.now:
            default = TimestampNow()

        self.default = default
        super().__init__(default=default, **kwargs)

    ###########################################################################
    # For update queries

    def __add__(self, value: timedelta) -> QueryString:
        return self.timedelta_delegate.get_querystring(
            column=self,
            operator="+",
            value=value,
            engine_type=self._meta.engine_type,
        )

    def __radd__(self, value: timedelta) -> QueryString:
        return self.__add__(value)

    def __sub__(self, value: timedelta) -> QueryString:
        return self.timedelta_delegate.get_querystring(
            column=self,
            operator="-",
            value=value,
            engine_type=self._meta.engine_type,
        )

    ###########################################################################
    # Descriptors

    @overload
    def __get__(self, obj: Table, objtype=None) -> datetime: ...

    @overload
    def __get__(self, obj: None, objtype=None) -> Timestamp: ...

    def __get__(self, obj, objtype=None):
        return obj.__dict__[self._meta.name] if obj else self

    def __set__(self, obj, value: Union[datetime, None]):
        obj.__dict__[self._meta.name] = value


class Timestamptz(Column):
    """
    Used for storing timezone aware datetimes. Uses the ``datetime`` type for
    values. The values are converted to UTC in the database, and are also
    returned as UTC.

    **Example**

    .. code-block:: python

        import datetime

        class Concert(Table):
            starts = Timestamptz()

        # Create
        >>> await Concert(
        ...    starts=datetime.datetime(
        ...        year=2050, month=1, day=1, tzinfo=datetime.timezone.tz
        ...    )
        ... ).save()

        # Query
        >>> await Concert.select(Concert.starts)
        {
            'starts': datetime.datetime(
                2050, 1, 1, 0, 0, tzinfo=datetime.timezone.utc
            )
        }

    """

    value_type = datetime

    # Currently just used by ModelBuilder, to know that we want a timezone
    # aware datetime.
    tz_aware = True

    timedelta_delegate = TimedeltaDelegate()

    def __init__(
        self,
        default: TimestamptzArg = TimestamptzNow(),
        **kwargs: Unpack[ColumnKwargs],
    ) -> None:
        self._validate_default(
            default, TimestamptzArg.__args__  # type: ignore
        )

        if isinstance(default, datetime):
            default = TimestamptzCustom.from_datetime(default)

        if default == datetime.now:
            default = TimestamptzNow()

        self.default = default
        super().__init__(default=default, **kwargs)

    ###########################################################################
    # For update queries

    def __add__(self, value: timedelta) -> QueryString:
        return self.timedelta_delegate.get_querystring(
            column=self,
            operator="+",
            value=value,
            engine_type=self._meta.engine_type,
        )

    def __radd__(self, value: timedelta) -> QueryString:
        return self.__add__(value)

    def __sub__(self, value: timedelta) -> QueryString:
        return self.timedelta_delegate.get_querystring(
            column=self,
            operator="-",
            value=value,
            engine_type=self._meta.engine_type,
        )

    ###########################################################################
    # Descriptors

    @overload
    def __get__(self, obj: Table, objtype=None) -> datetime: ...

    @overload
    def __get__(self, obj: None, objtype=None) -> Timestamptz: ...

    def __get__(self, obj, objtype=None):
        return obj.__dict__[self._meta.name] if obj else self

    def __set__(self, obj, value: Union[datetime, None]):
        obj.__dict__[self._meta.name] = value


class Date(Column):
    """
    Used for storing dates. Uses the ``date`` type for values.

    **Example**

    .. code-block:: python

        import datetime

        class Concert(Table):
            starts = Date()

        # Create
        >>> await Concert(
        ...     starts=datetime.date(year=2020, month=1, day=1)
        ... ).save()

        # Query
        >>> await Concert.select(Concert.starts)
        {'starts': datetime.date(2020, 1, 1)}

    """

    value_type = date
    timedelta_delegate = TimedeltaDelegate()

    def __init__(
        self,
        default: DateArg = DateNow(),
        **kwargs: Unpack[ColumnKwargs],
    ) -> None:
        self._validate_default(default, DateArg.__args__)  # type: ignore

        if isinstance(default, date):
            default = DateCustom.from_date(default)

        if default == date.today:
            default = DateNow()

        self.default = default
        super().__init__(default=default, **kwargs)

    ###########################################################################
    # For update queries

    def __add__(self, value: timedelta) -> QueryString:
        return self.timedelta_delegate.get_querystring(
            column=self,
            operator="+",
            value=value,
            engine_type=self._meta.engine_type,
        )

    def __radd__(self, value: timedelta) -> QueryString:
        return self.__add__(value)

    def __sub__(self, value: timedelta) -> QueryString:
        return self.timedelta_delegate.get_querystring(
            column=self,
            operator="-",
            value=value,
            engine_type=self._meta.engine_type,
        )

    ###########################################################################
    # Descriptors

    @overload
    def __get__(self, obj: Table, objtype=None) -> date: ...

    @overload
    def __get__(self, obj: None, objtype=None) -> Date: ...

    def __get__(self, obj, objtype=None):
        return obj.__dict__[self._meta.name] if obj else self

    def __set__(self, obj, value: Union[date, None]):
        obj.__dict__[self._meta.name] = value


class Time(Column):
    """
    Used for storing times. Uses the ``time`` type for values.

    **Example**

    .. code-block:: python

        import datetime

        class Concert(Table):
            starts = Time()

        # Create
        >>> await Concert(
        ...    starts=datetime.time(hour=20, minute=0, second=0)
        ... ).save()

        # Query
        >>> await Concert.select(Concert.starts)
        {'starts': datetime.time(20, 0, 0)}

    """

    value_type = time
    timedelta_delegate = TimedeltaDelegate()

    def __init__(
        self,
        default: TimeArg = TimeNow(),
        **kwargs: Unpack[ColumnKwargs],
    ) -> None:
        self._validate_default(default, TimeArg.__args__)  # type: ignore

        if isinstance(default, time):
            default = TimeCustom.from_time(default)

        self.default = default
        super().__init__(default=default, **kwargs)

    ###########################################################################
    # For update queries

    def __add__(self, value: timedelta) -> QueryString:
        return self.timedelta_delegate.get_querystring(
            column=self,
            operator="+",
            value=value,
            engine_type=self._meta.engine_type,
        )

    def __radd__(self, value: timedelta) -> QueryString:
        return self.__add__(value)

    def __sub__(self, value: timedelta) -> QueryString:
        return self.timedelta_delegate.get_querystring(
            column=self,
            operator="-",
            value=value,
            engine_type=self._meta.engine_type,
        )

    ###########################################################################
    # Descriptors

    @overload
    def __get__(self, obj: Table, objtype=None) -> time: ...

    @overload
    def __get__(self, obj: None, objtype=None) -> Time: ...

    def __get__(self, obj, objtype=None):
        return obj.__dict__[self._meta.name] if obj else self

    def __set__(self, obj, value: Union[time, None]):
        obj.__dict__[self._meta.name] = value


class Interval(Column):
    """
    Used for storing timedeltas. Uses the ``timedelta`` type for values.

    **Example**

    .. code-block:: python

        from datetime import timedelta

        class Concert(Table):
            duration = Interval()

        # Create
        >>> await Concert(
        ...    duration=timedelta(hours=2)
        ... ).save()

        # Query
        >>> await Concert.select(Concert.duration)
        {'duration': datetime.timedelta(seconds=7200)}

    """

    value_type = timedelta
    timedelta_delegate = TimedeltaDelegate()

    def __init__(
        self,
        default: IntervalArg = IntervalCustom(),
        **kwargs: Unpack[ColumnKwargs],
    ) -> None:
        self._validate_default(default, IntervalArg.__args__)  # type: ignore

        if isinstance(default, timedelta):
            default = IntervalCustom.from_timedelta(default)

        self.default = default
        super().__init__(default=default, **kwargs)

    @property
    def column_type(self):
        engine_type = self._meta.engine_type
        if engine_type in ("postgres", "cockroach"):
            return "INTERVAL"
        elif engine_type == "sqlite":
            # We can't use 'INTERVAL' because the type affinity in SQLite would
            # make it an integer - but we need a text field.
            # https://sqlite.org/datatype3.html#determination_of_column_affinity
            return "SECONDS"
        raise Exception("Unrecognized engine type")

    ###########################################################################
    # For update queries

    def __add__(self, value: timedelta) -> QueryString:
        return self.timedelta_delegate.get_querystring(
            column=self,
            operator="+",
            value=value,
            engine_type=self._meta.engine_type,
        )

    def __radd__(self, value: timedelta) -> QueryString:
        return self.__add__(value)

    def __sub__(self, value: timedelta) -> QueryString:
        return self.timedelta_delegate.get_querystring(
            column=self,
            operator="-",
            value=value,
            engine_type=self._meta.engine_type,
        )

    ###########################################################################
    # Descriptors

    @overload
    def __get__(self, obj: Table, objtype=None) -> timedelta: ...

    @overload
    def __get__(self, obj: None, objtype=None) -> Interval: ...

    def __get__(self, obj, objtype=None):
        return obj.__dict__[self._meta.name] if obj else self

    def __set__(self, obj, value: Union[timedelta, None]):
        obj.__dict__[self._meta.name] = value


###############################################################################


class Boolean(Column):
    """
    Used for storing ``True`` / ``False`` values. Uses the ``bool`` type for
    values.

    **Example**

    .. code-block:: python

        class Band(Table):
            has_drummer = Boolean()

        # Create
        >>> await Band(has_drummer=True).save()

        # Query
        >>> await Band.select(Band.has_drummer)
        {'has_drummer': True}

    """

    value_type = bool

    def __init__(
        self,
        default: Union[bool, Enum, Callable[[], bool], None] = False,
        **kwargs: Unpack[ColumnKwargs],
    ) -> None:
        self._validate_default(default, (bool, None))
        self.default = default
        super().__init__(default=default, **kwargs)

    def eq(self, value) -> Where:
        """
        When using ``Boolean`` columns in ``where`` clauses, some Python
        linters don't like it when you do something like:

        .. code-block:: python

            await MyTable.select().where(
                MyTable.some_boolean_column == True
            )

        It's more Pythonic to use ``is True`` rather than ``== True``, which is
        why linters complain. The work around is to do the following instead:

        .. code-block:: python

            await MyTable.select().where(
                MyTable.some_boolean_column.__eq__(True)
            )

        Using the ``__eq__`` magic method is a bit untidy, which is why this
        ``eq`` method exists.

        .. code-block:: python

            await MyTable.select().where(
                MyTable.some_boolean_column.eq(True)
            )

        The ``ne`` method exists for the same reason, for ``!=``.

        """
        return self.__eq__(value)

    def ne(self, value) -> Where:
        """
        See the ``eq`` method for more details.
        """
        return self.__ne__(value)

    ###########################################################################
    # Descriptors

    @overload
    def __get__(self, obj: Table, objtype=None) -> bool: ...

    @overload
    def __get__(self, obj: None, objtype=None) -> Boolean: ...

    def __get__(self, obj, objtype=None):
        return obj.__dict__[self._meta.name] if obj else self

    def __set__(self, obj, value: Union[bool, None]):
        obj.__dict__[self._meta.name] = value


###############################################################################


class Numeric(Column):
    """
    Used for storing decimal numbers, when precision is important. An example
    use case is storing financial data. The value is returned as a ``Decimal``.

    **Example**

    .. code-block:: python

        from decimal import Decimal

        class Ticket(Table):
            price = Numeric(digits=(5,2))

        # Create
        >>> await Ticket(price=Decimal('50.0')).save()

        # Query
        >>> await Ticket.select(Ticket.price)
        {'price': Decimal('50.0')}

    :param digits:
        When creating the column, you specify how many digits are allowed
        using a tuple. The first value is the ``precision``, which is the
        total number of digits allowed. The second value is the ``range``,
        which specifies how many of those digits are after the decimal
        point. For example, to store monetary values up to £999.99, the
        digits argument is ``(5,2)``.

    """

    value_type = decimal.Decimal

    @property
    def column_type(self):
        engine_type = self._meta.engine_type
        if engine_type == "cockroach":
            return "NUMERIC"  # All Numeric is the same for Cockroach.
        if self.digits:
            return f"NUMERIC({self.precision}, {self.scale})"
        else:
            return "NUMERIC"

    @property
    def precision(self) -> Optional[int]:
        """
        The total number of digits allowed.
        """
        return self.digits[0] if self.digits is not None else None

    @property
    def scale(self) -> Optional[int]:
        """
        The number of digits after the decimal point.
        """
        return self.digits[1] if self.digits is not None else None

    def __init__(
        self,
        digits: Optional[tuple[int, int]] = None,
        default: Union[
            decimal.Decimal, Enum, Callable[[], decimal.Decimal], None
        ] = decimal.Decimal(0.0),
        **kwargs: Unpack[ColumnKwargs],
    ) -> None:
        if isinstance(digits, tuple):
            if len(digits) != 2:
                raise ValueError(
                    "The `digits` argument should be a tuple of length 2, "
                    "with the first value being the precision, and the second "
                    "value being the scale."
                )
        elif digits is not None:
            raise ValueError("The digits argument should be a tuple.")

        self._validate_default(default, (decimal.Decimal, None))

        self.default = default
        self.digits = digits
        super().__init__(default=default, digits=digits, **kwargs)

    ###########################################################################
    # Descriptors

    @overload
    def __get__(self, obj: Table, objtype=None) -> decimal.Decimal: ...

    @overload
    def __get__(self, obj: None, objtype=None) -> Numeric: ...

    def __get__(self, obj, objtype=None):
        return obj.__dict__[self._meta.name] if obj else self

    def __set__(self, obj, value: Union[decimal.Decimal, None]):
        obj.__dict__[self._meta.name] = value


class Decimal(Numeric):
    """
    An alias for Numeric.
    """

    ###########################################################################
    # Descriptors

    @overload
    def __get__(self, obj: Table, objtype=None) -> decimal.Decimal: ...

    @overload
    def __get__(self, obj: None, objtype=None) -> Decimal: ...

    def __get__(self, obj, objtype=None):
        return obj.__dict__[self._meta.name] if obj else self

    def __set__(self, obj, value: Union[decimal.Decimal, None]):
        obj.__dict__[self._meta.name] = value


class Real(Column):
    """
    Can be used instead of ``Numeric`` for storing numbers, when precision
    isn't as important. The ``float`` type is used for values.

    **Example**

    .. code-block:: python

        class Concert(Table):
            rating = Real()

        # Create
        >>> await Concert(rating=7.8).save()

        # Query
        >>> await Concert.select(Concert.rating)
        {'rating': 7.8}

    """

    value_type = float

    def __init__(
        self,
        default: Union[float, Enum, Callable[[], float], None] = 0.0,
        **kwargs: Unpack[ColumnKwargs],
    ) -> None:
        if isinstance(default, int):
            # For example, allow `0` as a valid default.
            default = float(default)

        self._validate_default(default, (float, None))
        self.default = default
        super().__init__(default=default, **kwargs)

    ###########################################################################
    # Descriptors

    @overload
    def __get__(self, obj: Table, objtype=None) -> float: ...

    @overload
    def __get__(self, obj: None, objtype=None) -> Real: ...

    def __get__(self, obj, objtype=None):
        return obj.__dict__[self._meta.name] if obj else self

    def __set__(self, obj, value: Union[float, None]):
        obj.__dict__[self._meta.name] = value


class Float(Real):
    """
    An alias for Real.
    """

    ###########################################################################
    # Descriptors

    @overload
    def __get__(self, obj: Table, objtype=None) -> float: ...

    @overload
    def __get__(self, obj: None, objtype=None) -> Float: ...

    def __get__(self, obj, objtype=None):
        return obj.__dict__[self._meta.name] if obj else self

    def __set__(self, obj, value: Union[float, None]):
        obj.__dict__[self._meta.name] = value


class DoublePrecision(Real):
    """
    The same as ``Real``, except the numbers are stored with greater precision.
    """

    @property
    def column_type(self):
        return "DOUBLE PRECISION"

    ###########################################################################
    # Descriptors

    @overload
    def __get__(self, obj: Table, objtype=None) -> float: ...

    @overload
    def __get__(self, obj: None, objtype=None) -> DoublePrecision: ...

    def __get__(self, obj, objtype=None):
        return obj.__dict__[self._meta.name] if obj else self

    def __set__(self, obj, value: Union[float, None]):
        obj.__dict__[self._meta.name] = value


###############################################################################


@dataclass
class ForeignKeySetupResponse:
    is_lazy: bool


class ForeignKey(Column, Generic[ReferencedTable]):
    """
    Used to reference another table. Uses the same type as the primary key
    column on the table it references.

    **Example**

    .. code-block:: python

        class Band(Table):
            manager = ForeignKey(references=Manager)

        # Create
        >>> await Band(manager=1).save()

        # Query
        >>> await Band.select(Band.manager)
        {'manager': 1}

        # Query object
        >>> band = await Band.objects().first()
        >>> band.manager
        1

    **Joins**

    You also use it to perform joins:

    .. code-block:: python

        >>> await Band.select(Band.name, Band.manager.name).first()
        {'name': 'Pythonistas', 'manager.name': 'Guido'}

    To retrieve all of the columns in the related table:

    .. code-block:: python

        >>> await Band.select(Band.name, *Band.manager.all_columns()).first()
        {'name': 'Pythonistas', 'manager.id': 1, 'manager.name': 'Guido'}

    To get a referenced row as an object:

    .. code-block:: python

        manager = await Manager.objects().where(
            Manager.id == some_band.manager
        )

    Or use either of the following, which are just a proxy to the above:

    .. code-block:: python

        manager = await band.get_related('manager')
        manager = await band.get_related(Band.manager)

    To change the manager:

    .. code-block:: python

        band.manager = some_manager_id
        await band.save()

    :param references:
        The ``Table`` being referenced.

        .. code-block:: python

            class Band(Table):
                manager = ForeignKey(references=Manager)

        A table can have a reference to itself, if you pass a ``references``
        argument of ``'self'``.

        .. code-block:: python

            class Musician(Table):
                name = Varchar(length=100)
                instructor = ForeignKey(references='self')

        In certain situations, you may be unable to reference a ``Table`` class
        if it causes a circular dependency. Try and avoid these by refactoring
        your code. If unavoidable, you can specify a lazy reference. If the
        ``Table`` is defined in the same file:

        .. code-block:: python

            class Band(Table):
                manager = ForeignKey(references='Manager')

        If the ``Table`` is defined in a Piccolo app:

        .. code-block:: python

            from piccolo.columns.reference import LazyTableReference

            class Band(Table):
                manager = ForeignKey(
                    references=LazyTableReference(
                       table_class_name="Manager", app_name="my_app",
                    )
                )

        If you aren't using Piccolo apps, you can specify a ``Table`` in any
        Python module:

        .. code-block:: python

            from piccolo.columns.reference import LazyTableReference

            class Band(Table):
                manager = ForeignKey(
                    references=LazyTableReference(
                       table_class_name="Manager",
                       module_path="some_module.tables",
                    )
                    # Alternatively, Piccolo will interpret this string as
                    # the same as above:
                    # references="some_module.tables.Manager"
                )

    :param on_delete:
        Determines what the database should do when a row is deleted with
        foreign keys referencing it. If set to ``OnDelete.cascade``, any rows
        referencing the deleted row are also deleted.

        Options:

        * ``OnDelete.cascade`` (default)
        * ``OnDelete.restrict``
        * ``OnDelete.no_action``
        * ``OnDelete.set_null``
        * ``OnDelete.set_default``

        To learn more about the different options, see the `Postgres docs <https://www.postgresql.org/docs/current/ddl-constraints.html#DDL-CONSTRAINTS-FK>`_.

        .. code-block:: python

            from piccolo.columns import OnDelete

            class Band(Table):
                name = ForeignKey(
                    references=Manager,
                    on_delete=OnDelete.cascade
                )

    :param on_update:
        Determines what the database should do when a row has it's primary key
        updated. If set to ``OnUpdate.cascade``, any rows referencing the
        updated row will have their references updated to point to the new
        primary key.

        Options:

        * ``OnUpdate.cascade`` (default)
        * ``OnUpdate.restrict``
        * ``OnUpdate.no_action``
        * ``OnUpdate.set_null``
        * ``OnUpdate.set_default``

        To learn more about the different options, see the `Postgres docs <https://www.postgresql.org/docs/current/ddl-constraints.html#DDL-CONSTRAINTS-FK>`_.

        .. code-block:: python

            from piccolo.columns import OnUpdate

            class Band(Table):
                name = ForeignKey(
                    references=Manager,
                    on_update=OnUpdate.cascade
                )

    :param target_column:
        By default the ``ForeignKey`` references the primary key column on the
        related table. You can specify an alternative column (it must have a
        unique constraint on it though). For example:

        .. code-block:: python

            # Passing in a column reference:
            ForeignKey(references=Manager, target_column=Manager.passport_number)

            # Or just the column name:
            ForeignKey(references=Manager, target_column='passport_number')

    """  # noqa: E501

    _foreign_key_meta: ForeignKeyMeta

    @property
    def column_type(self):
        """
        A ``ForeignKey`` column needs to have the same type as the primary key
        column of the table being referenced.
        """
        target_column = self._foreign_key_meta.resolved_target_column

        if isinstance(target_column, BigSerial):
            return BigInt()._get_column_type(
                engine_type=self._meta.engine_type
            )
        elif isinstance(target_column, Serial):
            return Integer().column_type
        else:
            return target_column.column_type

    @property
    def value_type(self):
        """
        The value type matches that of the primary key being referenced.
        """
        target_column = self._foreign_key_meta.resolved_target_column
        return target_column.value_type

    @overload
    def __init__(
        self,
        references: type[ReferencedTable],
        default: Any = None,
        null: bool = True,
        on_delete: OnDelete = OnDelete.cascade,
        on_update: OnUpdate = OnUpdate.cascade,
        target_column: Union[str, Column, None] = None,
        **kwargs,
    ) -> None: ...

    @overload
    def __init__(
        self,
        references: LazyTableReference,
        default: Any = None,
        null: bool = True,
        on_delete: OnDelete = OnDelete.cascade,
        on_update: OnUpdate = OnUpdate.cascade,
        target_column: Union[str, Column, None] = None,
        **kwargs,
    ) -> None: ...

    @overload
    def __init__(
        self,
        references: str,
        default: Any = None,
        null: bool = True,
        on_delete: OnDelete = OnDelete.cascade,
        on_update: OnUpdate = OnUpdate.cascade,
        target_column: Union[str, Column, None] = None,
        **kwargs,
    ) -> None: ...

    def __init__(
        self,
        references: Union[type[ReferencedTable], LazyTableReference, str],
        default: Any = None,
        null: bool = True,
        on_delete: OnDelete = OnDelete.cascade,
        on_update: OnUpdate = OnUpdate.cascade,
        target_column: Union[str, Column, None] = None,
        **kwargs,
    ) -> None:
        from piccolo.table import Table

        if inspect.isclass(references):
            if issubclass(references, Table):
                # Using this to validate the default value - will raise a
                # ValueError if incorrect.
                if isinstance(references._meta.primary_key, Serial):
                    Integer(default=default, null=null)
                else:
                    references._meta.primary_key.__class__(
                        default=default, null=null
                    )

        self.default = default

        kwargs.update(
            {
                "references": references,
                "on_delete": on_delete,
                "on_update": on_update,
                "null": null,
                "target_column": target_column,
            }
        )

        super().__init__(**kwargs)

        # The ``TableMetaclass``` sets the actual value for
        # ``ForeignKeyMeta.references``, if the user passed in a string.
        self._foreign_key_meta = ForeignKeyMeta(
            references=Table if isinstance(references, str) else references,
            on_delete=on_delete,
            on_update=on_update,
            target_column=target_column,
        )

    def _setup(self, table_class: type[Table]) -> ForeignKeySetupResponse:
        """
        This is called by the ``TableMetaclass``. A ``ForeignKey`` column can
        only be completely setup once it's parent ``Table`` is known.

        :param table_class:
            The parent ``Table`` class for this column.

        """
        from piccolo.table import Table

        params = self._meta.params
        references = params["references"]

        if isinstance(references, str):
            if references == "self":
                references = table_class
            else:
                if "." in references:
                    # Don't allow relative modules - this may change in
                    # the future.
                    if references.startswith("."):
                        raise ValueError("Relative imports aren't allowed")

                    module_path, table_class_name = references.rsplit(
                        ".", maxsplit=1
                    )
                else:
                    table_class_name = references
                    module_path = table_class.__module__

                references = LazyTableReference(
                    table_class_name=table_class_name,
                    module_path=module_path,
                )

        is_lazy = isinstance(references, LazyTableReference)
        is_table_class = inspect.isclass(references) and issubclass(
            references, Table
        )

        if is_lazy or is_table_class:
            self._foreign_key_meta.references = references
        else:
            raise ValueError(
                "Error - ``references`` must be a ``Table`` subclass, or "
                "a ``LazyTableReference`` instance."
            )

        if is_table_class:
            # Record the reverse relationship on the target table.
            cast(type[Table], references)._meta._foreign_key_references.append(
                self
            )

            # Allow columns on the referenced table to be accessed via
            # auto completion.
            self.set_proxy_columns()

        return ForeignKeySetupResponse(is_lazy=is_lazy)

    def copy(self) -> ForeignKey:
        column: ForeignKey = copy.copy(self)
        column._meta = self._meta.copy()
        column._foreign_key_meta = self._foreign_key_meta.copy()
        return column

    def all_columns(
        self, exclude: Optional[list[Union[Column, str]]] = None
    ) -> list[Column]:
        """
        Allow a user to access all of the columns on the related table. This is
        intended for use with ``select`` queries, and saves the user from
        typing out all of the columns by hand.

        For example:

        .. code-block:: python

            await Band.select(Band.name, Band.manager.all_columns())

            # Equivalent to:
            await Band.select(
                Band.name,
                Band.manager.id,
                Band.manager.name
            )

        To exclude certain columns:

        .. code-block:: python

            await Band.select(
                Band.name,
                Band.manager.all_columns(
                    exclude=[Band.manager.id]
                )
            )

        :param exclude:
            Columns to exclude - can be the name of a column, or a column
            instance. For example ``['id']`` or ``[Band.manager.id]``.

        """
        if exclude is None:
            exclude = []
        _fk_meta = object.__getattribute__(self, "_foreign_key_meta")

        excluded_column_names = [
            i._meta.name if isinstance(i, Column) else i for i in exclude
        ]

        return [
            getattr(self, column._meta.name)
            for column in _fk_meta.resolved_references._meta.columns
            if column._meta.name not in excluded_column_names
        ]

    def reverse(self) -> ForeignKey:
        """
        If there's a unique foreign key, this function reverses it.

        .. code-block:: python

            class Band(Table):
                name = Varchar()

            class FanClub(Table):
                band = ForeignKey(Band, unique=True)
                address = Text()

            class Treasurer(Table):
                fan_club = ForeignKey(FanClub, unique=True)
                name = Varchar()

        It's helpful with ``get_related``, for example:

        .. code-block:: python

            >>> band = await Band.objects().first()
            >>> await band.get_related(FanClub.band.reverse())
            <Fan Club: 1>

        It works multiple levels deep:

        .. code-block:: python

            >>> await band.get_related(Treasurer.fan_club._.band.reverse())
            <Treasurer: 1>

        """
        if not self._meta.unique or any(
            not i._meta.unique for i in self._meta.call_chain
        ):
            raise ValueError("Only reverse unique foreign keys.")

        foreign_keys = [*self._meta.call_chain, self]

        root_foreign_key = foreign_keys[0]
        target_column = (
            root_foreign_key._foreign_key_meta.resolved_target_column
        )
        foreign_key = target_column.join_on(root_foreign_key)

        call_chain = []
        for fk in reversed(foreign_keys[1:]):
            target_column = fk._foreign_key_meta.resolved_target_column
            call_chain.append(target_column.join_on(fk))

        foreign_key._meta.call_chain = call_chain

        return foreign_key

    def all_related(
        self, exclude: Optional[list[Union[ForeignKey, str]]] = None
    ) -> list[ForeignKey]:
        """
        Returns each ``ForeignKey`` column on the related table. This is
        intended for use with ``objects`` queries, where you want to return
        all of the related tables as nested objects.

        For example:

        .. code-block:: python

            class Band(Table):
                name = Varchar()

            class Concert(Table):
                name = Varchar()
                band_1 = ForeignKey(Band)
                band_2 = ForeignKey(Band)

            class Tour(Table):
                name = Varchar()
                concert = ForeignKey(Concert)

            await Tour.objects(Tour.concert, Tour.concert.all_related())

            # Equivalent to
            await Tour.objects(
                Tour.concert,
                Tour.concert.band_1,
                Tour.concert.band_2
            )

        :param exclude:
            Columns to exclude - can be the name of a column, or a
            ``ForeignKey`` instance. For example ``['band_1']`` or
            ``[Tour.concert.band_1]``.

        """
        if exclude is None:
            exclude = []
        _fk_meta: ForeignKeyMeta = object.__getattribute__(
            self, "_foreign_key_meta"
        )
        related_fk_columns = (
            _fk_meta.resolved_references._meta.foreign_key_columns
        )
        excluded_column_names = [
            i._meta.name if isinstance(i, ForeignKey) else i for i in exclude
        ]
        return [
            getattr(self, fk_column._meta.name)
            for fk_column in related_fk_columns
            if fk_column._meta.name not in excluded_column_names
        ]

    def set_proxy_columns(self) -> None:
        """
        In order to allow a fluent interface, where tables can be traversed
        using ForeignKeys (e.g. ``Band.manager.name``), we add attributes to
        the ``ForeignKey`` column for each column in the table being pointed
        to.
        """
        _fk_meta = object.__getattribute__(self, "_foreign_key_meta")
        for column in _fk_meta.resolved_references._meta.columns:
            _column: Column = column.copy()
            setattr(self, _column._meta.name, _column)
            _fk_meta.proxy_columns.append(_column)

    @property
    def _(self) -> type[ReferencedTable]:
        """
        This allows us specify joins in a way which is friendly to static type
        checkers like Mypy and Pyright.

        Whilst this works::

            # Fetch the band's name, and their manager's name via a foreign
            # key:
            await Band.select(Band.name, Band.manager.name)

        There currently isn't a 100% reliable way to tell static type checkers
        that ``Band.manager.name`` refers to a ``name`` column on the
        ``Manager`` table.

        However, by using the ``_`` property, it works perfectly. Instead
        of ``Band.manager.name`` we use ``Band.manager._.name``::

            await Band.select(Band.name, Band.manager._.name)

        So when doing joins, after every foreign key we use ``._.`` instead of
        ``.``. An easy way to remember this is ``._.`` looks a bit like a
        connector in a diagram.

        As Python's typing support increases, we'd love ``Band.manager.name``
        to have the same static typing as ``Band.manager._.name`` (using some
        kind of ``Proxy`` type), but for now this is the best solution, and is
        a huge improvement in developer experience, as static type checkers
        easily know if any of your joins contain typos.

        """
        return cast(type[ReferencedTable], self)

    def __getattribute__(self, name: str) -> Union[Column, Any]:
        """
        Returns attributes unmodified unless they're Column instances, in which
        case a copy is returned with an updated call_chain (which records the
        joins required).
        """
        # If the ForeignKey is using a lazy reference, we need to set the
        # attributes here. Attributes starting with an underscore are
        # unlikely to be column names.
        if not name.startswith("_") and name not in dir(self):
            try:
                _foreign_key_meta = object.__getattribute__(
                    self, "_foreign_key_meta"
                )
            except AttributeError:
                pass
            else:
                if _foreign_key_meta.proxy_columns == [] and isinstance(
                    _foreign_key_meta.references, LazyTableReference
                ):
                    object.__getattribute__(self, "set_proxy_columns")()

        value = object.__getattribute__(self, name)

        if name.startswith("_"):
            return value

        foreignkey_class: type[ForeignKey] = object.__getattribute__(
            self, "__class__"
        )

        if isinstance(value, foreignkey_class):  # i.e. a ForeignKey
            new_column = value.copy()
            new_column._meta.call_chain.append(self)

            # We have to set limits to the call chain because Table 1 can
            # reference Table 2, which references Table 1, creating an endless
            # loop. For now an arbitrary limit is set of 10 levels deep.
            # When querying a call chain more than 10 levels deep, an error
            # will be raised. Often there are more effective ways of
            # structuring a query than joining so many tables anyway.
            if len(new_column._meta.call_chain) >= 10:
                raise Exception("Call chain too long!")

            foreign_key_meta: ForeignKeyMeta = object.__getattribute__(
                new_column, "_foreign_key_meta"
            )

            for proxy_column in foreign_key_meta.proxy_columns:
                try:
                    delattr(new_column, proxy_column._meta.name)
                except Exception:
                    pass

            foreign_key_meta.proxy_columns = []

            for (
                column
            ) in value._foreign_key_meta.resolved_references._meta.columns:
                _column: Column = column.copy()
                _column._meta.call_chain = list(new_column._meta.call_chain)
                setattr(new_column, _column._meta.name, _column)
                foreign_key_meta.proxy_columns.append(_column)

            return new_column
        elif issubclass(type(value), Column):
            new_column = value.copy()

            column_meta: ColumnMeta = object.__getattribute__(self, "_meta")

            new_column._meta.call_chain = column_meta.call_chain.copy()

            new_column._meta.call_chain.append(self)
            return new_column
        else:
            return value

    ###########################################################################
    # Descriptors

    @overload
    def __get__(self, obj: Table, objtype=None) -> Any: ...

    @overload
    def __get__(
        self, obj: None, objtype=None
    ) -> ForeignKey[ReferencedTable]: ...

    @overload
    def __get__(self, obj: Any, objtype=None) -> Any: ...

    def __get__(self, obj, objtype=None):
        return obj.__dict__[self._meta.name] if obj else self

    def __set__(self, obj, value: Any):
        obj.__dict__[self._meta.name] = value


###############################################################################


class JSON(Column):
    """
    Used for storing JSON strings. The data is stored as text. This can be
    preferable to JSONB if you just want to store and retrieve JSON without
    querying it directly. It works with SQLite and Postgres.

    :param default:
        Either a JSON string can be provided, or a Python ``dict`` or ``list``
        which is then converted to a JSON string.

    """

    value_type = str

    def __init__(
        self,
        default: Union[
            str,
            list,
            dict,
            Callable[[], Union[str, list, dict]],
            None,
        ] = "{}",
        **kwargs: Unpack[ColumnKwargs],
    ) -> None:
        self._validate_default(default, (str, list, dict, None))

        if isinstance(default, (list, dict)):
            default = dump_json(default)

        self.default = default
        super().__init__(default=default, **kwargs)

        self.json_operator: Optional[str] = None

    @property
    def column_type(self):
        engine_type = self._meta.engine_type
        if engine_type == "cockroach":
            return "JSONB"  # Cockroach is always JSONB.
        else:
            return "JSON"

    ###########################################################################

    def arrow(self, key: Union[str, int, QueryString]) -> GetChildElement:
        """
        Allows a child element of the JSON structure to be returned - for
        example::

            >>> await RecordingStudio.select(
            ...     RecordingStudio.facilities.arrow("restaurant")
            ... )

        """
        from piccolo.query.operators.json import GetChildElement

        alias = self._alias or self._meta.get_default_alias()
        return GetChildElement(identifier=self, key=key, alias=alias)

    def __getitem__(
        self, value: Union[str, int, QueryString]
    ) -> GetChildElement:
        """
        A shortcut for the ``arrow`` method, used for retrieving a child
        element.

        For example:

        .. code-block:: python

            >>> await RecordingStudio.select(
            ...     RecordingStudio.facilities["restaurant"]
            ... )

        """
        return self.arrow(key=value)

    def from_path(
        self,
        path: list[Union[str, int]],
    ) -> GetElementFromPath:
        """
        Allows an element of the JSON structure to be returned, which can be
        arbitrarily deep. For example::

            >>> await RecordingStudio.select(
            ...     RecordingStudio.facilities.from_path([
            ...         "technician",
            ...         0,
            ...         "first_name"
            ...     ])
            ... )

        It's the same as calling ``arrow`` multiple times, but is more
        efficient / convenient if extracting highly nested data::

            >>> await RecordingStudio.select(
            ...     RecordingStudio.facilities.arrow(
            ...         "technician"
            ...     ).arrow(
            ...         0
            ...     ).arrow(
            ...         "first_name"
            ...     )
            ... )

        """
        from piccolo.query.operators.json import GetElementFromPath

        alias = self._alias or self._meta.get_default_alias()
        return GetElementFromPath(identifier=self, path=path, alias=alias)

    ###########################################################################
    # Descriptors

    @overload
    def __get__(self, obj: Table, objtype=None) -> str: ...

    @overload
    def __get__(self, obj: None, objtype=None) -> JSON: ...

    def __get__(self, obj, objtype=None):
        return obj.__dict__[self._meta.name] if obj else self

    def __set__(self, obj, value: Union[str, dict]):
        obj.__dict__[self._meta.name] = value


class JSONB(JSON):
    """
    Used for storing JSON strings - Postgres / CochroachDB only. The data is
    stored in a binary format, and can be queried more efficiently. Insertion
    can be slower (as it needs to be converted to the binary format). The
    benefits of JSONB generally outweigh the downsides.

    :param default:
        Either a JSON string can be provided, or a Python ``dict`` or ``list``
        which is then converted to a JSON string.

    """

    @property
    def column_type(self):
        return "JSONB"  # Must be defined, we override column_type() in JSON()

    ###########################################################################
    # Descriptors

    @overload
    def __get__(self, obj: Table, objtype=None) -> str: ...

    @overload
    def __get__(self, obj: None, objtype=None) -> JSONB: ...

    def __get__(self, obj, objtype=None):
        return obj.__dict__[self._meta.name] if obj else self

    def __set__(self, obj, value: Union[str, dict]):
        obj.__dict__[self._meta.name] = value


###############################################################################


class Bytea(Column):
    """
    Used for storing bytes.

    **Example**

    .. code-block:: python

        class Token(Table):
            token = Bytea(default=b'token123')

        # Create
        >>> await Token(token=b'my-token').save()

        # Query
        >>> await Token.select(Token.token)
        {'token': b'my-token'}

    """

    value_type = bytes

    @property
    def column_type(self):
        engine_type = self._meta.engine_type
        if engine_type in ("postgres", "cockroach"):
            return "BYTEA"
        elif engine_type == "sqlite":
            return "BLOB"
        raise Exception("Unrecognized engine type")

    def __init__(
        self,
        default: Union[
            bytes,
            bytearray,
            Enum,
            Callable[[], bytes],
            Callable[[], bytearray],
            None,
        ] = b"",
        **kwargs: Unpack[ColumnKwargs],
    ) -> None:
        self._validate_default(default, (bytes, bytearray, None))

        if isinstance(default, bytearray):
            default = bytes(default)

        self.default = default
        super().__init__(default=default, **kwargs)

    ###########################################################################
    # Descriptors

    @overload
    def __get__(self, obj: Table, objtype=None) -> bytes: ...

    @overload
    def __get__(self, obj: None, objtype=None) -> Bytea: ...

    def __get__(self, obj, objtype=None):
        return obj.__dict__[self._meta.name] if obj else self

    def __set__(self, obj, value: bytes):
        obj.__dict__[self._meta.name] = value


class Blob(Bytea):
    """
    An alias for Bytea.
    """

    ###########################################################################
    # Descriptors

    @overload
    def __get__(self, obj: Table, objtype=None) -> bytes: ...

    @overload
    def __get__(self, obj: None, objtype=None) -> Blob: ...

    def __get__(self, obj, objtype=None):
        return obj.__dict__[self._meta.name] if obj else self

    def __set__(self, obj, value: bytes):
        obj.__dict__[self._meta.name] = value


###############################################################################


class ListProxy:
    """
    Sphinx's autodoc fails if we have this function signature::

        class Array(Column):

            def __init__(default=list):
                ...

    We can't use ``list`` as a default value without breaking autodoc (it
    doesn't seem to like it when a class type is used as a default), so
    instead we assign an instance of this class. It keeps both autodoc and MyPy
    happy. In ``Array.__init__`` we then swap it out for ``list``.
    """

    def __call__(self):
        return []

    def __repr__(self):
        return "list"


class Array(Column):
    """
    Used for storing lists of data.

    **Example**

    .. code-block:: python

        class Ticket(Table):
            seat_numbers = Array(base_column=Integer())

        # Create
        >>> await Ticket(seat_numbers=[34, 35, 36]).save()

        # Query
        >>> await Ticket.select(Ticket.seat_numbers)
        {'seat_numbers': [34, 35, 36]}

    """

    value_type = list

    def __init__(
        self,
        base_column: Column,
        default: Union[list, Enum, Callable[[], list], None] = ListProxy(),
        **kwargs: Unpack[ColumnKwargs],
    ) -> None:
        if isinstance(base_column, ForeignKey):
            raise ValueError("Arrays of ForeignKeys aren't allowed.")

        # This is a workaround because having `list` as a default breaks
        # Sphinx's autodoc.
        if isinstance(default, ListProxy):
            default = list

        self._validate_default(default, (list, None))

        choices = kwargs.get("choices")
        if choices is not None:
            self._validate_choices(
                choices, allowed_type=base_column.value_type
            )
            self._validated_choices = True

        # Usually columns are given a name by the Table metaclass, but in this
        # case we have to assign one manually to the base column.
        base_column._meta._name = base_column.__class__.__name__

        self.base_column = base_column
        self.default = default
        self.index: Optional[int] = None
        super().__init__(default=default, base_column=base_column, **kwargs)

    @property
    def column_type(self):
        engine_type = self._meta.engine_type
        if engine_type in ("postgres", "cockroach"):
            return f"{self.base_column.column_type}[]"
        elif engine_type == "sqlite":
            inner_column = self._get_inner_column()
            return (
                f"ARRAY_{inner_column.column_type}"
                if isinstance(
                    inner_column, (Date, Timestamp, Timestamptz, Time)
                )
                else "ARRAY"
            )
        raise Exception("Unrecognized engine type")

    def _setup_base_column(self, table_class: type[Table]):
        """
        Called from the ``Table.__init_subclass__`` - makes sure
        that the ``base_column`` has a reference to the parent table.
        """
        self.base_column._meta._table = table_class
        if isinstance(self.base_column, Array):
            self.base_column._setup_base_column(table_class=table_class)

    def _get_dimensions(self, start: int = 0) -> int:
        """
        A helper function to get the number of dimensions for the array. For
        example::

            >>> Array(Varchar())._get_dimensions()
            1

            >>> Array(Array(Varchar()))._get_dimensions()
            2

        :param start:
            Ignore this - it's just used for  calling this method recursively.

        """
        if isinstance(self.base_column, Array):
            return self.base_column._get_dimensions(start=start + 1)
        else:
            return start + 1

    def _get_inner_column(self) -> Column:
        """
        A helper function to get the innermost ``Column`` for the array. For
        example::

            >>> Array(Varchar())._get_inner_column()
            Varchar

            >>> Array(Array(Varchar()))._get_inner_column()
            Varchar

        """
        if isinstance(self.base_column, Array):
            return self.base_column._get_inner_column()
        else:
            return self.base_column

    def _get_inner_value_type(self) -> type:
        """
        A helper function to get the innermost value type for the array. For
        example::

            >>> Array(Varchar())._get_inner_value_type()
            str

            >>> Array(Array(Varchar()))._get_inner_value_type()
            str

        """
        return self._get_inner_column().value_type

    def __getitem__(self, value: int) -> Array:
        """
        Allows queries which retrieve an item from the array. The index starts
        with 0 for the first value. If you were to write the SQL by hand, the
        first index would be 1 instead (see `Postgres array docs <https://www.postgresql.org/docs/current/arrays.html>`_).

        However, we keep the first index as 0 to fit better with Python.

        For example:

        .. code-block:: python

            >>> await Ticket.select(Ticket.seat_numbers[0]).first()
            {'seat_numbers': 325}


        """  # noqa: E501
        engine_type = self._meta.engine_type
        if engine_type != "postgres" and engine_type != "cockroach":
            raise ValueError(
                "Only Postgres and Cockroach support array indexing."
            )

        if isinstance(value, int):
            if value < 0:
                raise ValueError("Only positive integers are allowed.")

            instance = cast(Array, self.copy())

            # We deliberately add 1, as Postgres treats the first array element
            # as index 1.
            instance.index = value + 1
            return instance
        else:
            raise ValueError("Only integers can be used for indexing.")

    def get_select_string(
        self, engine_type: str, with_alias=True
    ) -> QueryString:
        select_string = self._meta.get_full_name(with_alias=False)

        if isinstance(self.index, int):
            select_string += f"[{self.index}]"

        if with_alias:
            alias = self._alias or self._meta.get_default_alias()
            select_string += f' AS "{alias}"'

        return QueryString(select_string)

    def any(self, value: Any) -> Where:
        """
        Check if any of the items in the array match the given value.

        .. code-block:: python

            >>> await Ticket.select().where(Ticket.seat_numbers.any(510))

        """
        engine_type = self._meta.engine_type

        if engine_type in ("postgres", "cockroach"):
            return Where(column=self, value=value, operator=ArrayAny)
        elif engine_type == "sqlite":
            return self.like(f"%{value}%")
        else:
            raise ValueError("Unrecognised engine type")

    def not_any(self, value: Any) -> Where:
        """
        Check if the given value isn't in the array.

        .. code-block:: python

            >>> await Ticket.select().where(Ticket.seat_numbers.not_any(510))

        """
        engine_type = self._meta.engine_type

        if engine_type in ("postgres", "cockroach"):
            return Where(column=self, value=value, operator=ArrayNotAny)
        elif engine_type == "sqlite":
            return self.not_like(f"%{value}%")
        else:
            raise ValueError("Unrecognised engine type")

    def all(self, value: Any) -> Where:
        """
        Check if all of the items in the array match the given value.

        .. code-block:: python

            >>> await Ticket.select().where(Ticket.seat_numbers.all(510))

        """
        engine_type = self._meta.engine_type

        if engine_type in ("postgres", "cockroach"):
            return Where(column=self, value=value, operator=ArrayAll)
        elif engine_type == "sqlite":
            raise ValueError("Unsupported by SQLite")
        else:
            raise ValueError("Unrecognised engine type")

    def cat(self, value: ArrayType) -> QueryString:
        """
        A convenient way of accessing the
        :class:`ArrayCat <piccolo.query.functions.array.ArrayCat>` function.

        Used in an ``update`` query to concatenate two arrays.

        .. code-block:: python

            >>> await Ticket.update({
            ...     Ticket.seat_numbers: Ticket.seat_numbers.cat([1000])
            ... }).where(Ticket.id == 1)

        You can also use the ``+`` symbol if you prefer. To concatenate to
        the end:

        .. code-block:: python

            >>> await Ticket.update({
            ...     Ticket.seat_numbers: Ticket.seat_numbers + [1000]
            ... }).where(Ticket.id == 1)

        To concatenate to the start:

        .. code-block:: python

            >>> await Ticket.update({
            ...     Ticket.seat_numbers: [1000] + Ticket.seat_numbers
            ... }).where(Ticket.id == 1)

        You can concatenate multiple arrays in one go:

        .. code-block:: python

            >>> await Ticket.update({
            ...     Ticket.seat_numbers: [1000] + Ticket.seat_numbers + [2000]
            ... }).where(Ticket.id == 1)

        .. note:: Postgres / CockroachDB only

        """
        from piccolo.query.functions.array import ArrayCat

        # Keep this for backwards compatibility - we had this as a convenience
        # for users, but it would be nice to remove it in the future.
        if not isinstance(value, list):
            value = [value]

        return ArrayCat(array_1=self, array_2=value)

    def remove(self, value: ArrayItemType) -> QueryString:
        """
        A convenient way of accessing the
        :class:`ArrayRemove <piccolo.query.functions.array.ArrayRemove>`
        function.

        Used in an ``update`` query to remove an item from an array.

        .. code-block:: python

            >>> await Ticket.update({
            ...     Ticket.seat_numbers: Ticket.seat_numbers.remove(1000)
            ... }).where(Ticket.id == 1)

        .. note:: Postgres / CockroachDB only

        """
        from piccolo.query.functions.array import ArrayRemove

        return ArrayRemove(array=self, value=value)

    def prepend(self, value: ArrayItemType) -> QueryString:
        """
        A convenient way of accessing the
        :class:`ArrayPrepend <piccolo.query.functions.array.ArrayPrepend>`
        function.

        Used in an ``update`` query to prepend an item to an array.

        .. code-block:: python

            >>> await Ticket.update({
            ...     Ticket.seat_numbers: Ticket.seat_numbers.prepend(1000)
            ... }).where(Ticket.id == 1)

        .. note:: Postgres / CockroachDB only

        """
        from piccolo.query.functions.array import ArrayPrepend

        return ArrayPrepend(array=self, value=value)

    def append(self, value: ArrayItemType) -> QueryString:
        """
        A convenient way of accessing the
        :class:`ArrayAppend <piccolo.query.functions.array.ArrayAppend>`
        function.

        Used in an ``update`` query to append an item to an array.

        .. code-block:: python

            >>> await Ticket.update({
            ...     Ticket.seat_numbers: Ticket.seat_numbers.append(1000)
            ... }).where(Ticket.id == 1)

        .. note:: Postgres / CockroachDB only

        """
        from piccolo.query.functions.array import ArrayAppend

        return ArrayAppend(array=self, value=value)

    def replace(
        self, old_value: ArrayItemType, new_value: ArrayItemType
    ) -> QueryString:
        """
        A convenient way of accessing the
        :class:`ArrayReplace <piccolo.query.functions.array.ArrayReplace>`
        function.

        Used in an ``update`` query to replace each array item
        equal to the given value with a new value.

        .. code-block:: python

            >>> await Ticket.update({
            ...     Ticket.seat_numbers: Ticket.seat_numbers.replace(1000, 500)
            ... }).where(Ticket.id == 1)

        .. note:: Postgres / CockroachDB only

        """
        from piccolo.query.functions.array import ArrayReplace

        return ArrayReplace(self, old_value=old_value, new_value=new_value)

    def __add__(self, value: ArrayType) -> QueryString:
        return self.cat(value)

    def __radd__(self, value: ArrayType) -> QueryString:
        from piccolo.query.functions.array import ArrayCat

        return ArrayCat(array_1=value, array_2=self)

    ###########################################################################
    # Descriptors

    @overload
    def __get__(self, obj: Table, objtype=None) -> list[Any]: ...

    @overload
    def __get__(self, obj: None, objtype=None) -> Array: ...

    def __get__(self, obj, objtype=None):
        return obj.__dict__[self._meta.name] if obj else self

    def __set__(self, obj, value: list[Any]):
        obj.__dict__[self._meta.name] = value



from __future__ import annotations

import datetime
import os
import string
from dataclasses import dataclass
from itertools import chain
from types import ModuleType
from typing import Optional

import black
import jinja2

from piccolo import __VERSION__
from piccolo.apps.migrations.auto import (
    AlterStatements,
    DiffableTable,
    SchemaDiffer,
    SchemaSnapshot,
)
from piccolo.conf.apps import AppConfig, Finder
from piccolo.engine import SQLiteEngine
from piccolo.utils.printing import print_heading
from piccolo.utils.warnings import colored_warning

from .base import BaseMigrationManager

TEMPLATE_DIRECTORY = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "templates"
)

JINJA_ENV = jinja2.Environment(
    loader=jinja2.FileSystemLoader(searchpath=TEMPLATE_DIRECTORY),
)

MIGRATION_MODULES: dict[str, ModuleType] = {}

VALID_PYTHON_MODULE_CHARACTERS = string.ascii_lowercase + string.digits + "_"


def render_template(**kwargs):
    template = JINJA_ENV.get_template("migration.py.jinja")
    return template.render(version=__VERSION__, **kwargs)


def _create_migrations_folder(migrations_path: str) -> bool:
    """
    Creates the folder that migrations live in. Returns True/False depending
    on whether it was created or not.
    """
    if os.path.exists(migrations_path):
        return False
    os.mkdir(migrations_path)
    with open(os.path.join(migrations_path, "__init__.py"), "w"):
        pass
    return True


@dataclass
class NewMigrationMeta:
    migration_id: str
    migration_filename: str
    migration_path: str


def now():
    """
    In a separate function so it's easier to patch in tests.
    """
    return datetime.datetime.now()


def _generate_migration_meta(app_config: AppConfig) -> NewMigrationMeta:
    """
    Generates the migration ID and filename.
    """
    # The microseconds originally weren't part of the ID, but there was a
    # chance that the IDs would clash if the migrations are generated
    # programatically in quick succession (e.g. in a unit test), so they had
    # to be added. The trade off is a longer ID.
    _id = now().strftime("%Y-%m-%dT%H:%M:%S:%f")

    # Originally we just used the _id as the filename, but colons aren't
    # supported in Windows, so we need to sanitize it. We don't want to
    # change the _id format though, as it would break existing migrations.
    # The filename doesn't have any special significance - only the id matters.
    cleaned_id = _id.replace(":", "_").replace("-", "_").lower()

    # Just in case the app name contains characters which aren't valid for
    # a Python module.
    cleaned_app_name = "".join(
        [
            i
            for i in app_config.app_name.lower().replace("-", "_")
            if i in VALID_PYTHON_MODULE_CHARACTERS
        ]
    )

    filename = f"{cleaned_app_name}_{cleaned_id}"

    path = os.path.join(
        app_config.resolved_migrations_folder_path, f"{filename}.py"
    )

    return NewMigrationMeta(
        migration_id=_id, migration_filename=filename, migration_path=path
    )


class NoChanges(Exception):
    pass


async def _create_new_migration(
    app_config: AppConfig,
    auto: bool = False,
    description: str = "",
    auto_input: Optional[str] = None,
) -> NewMigrationMeta:
    """
    Creates a new migration file on disk.
    """
    meta = _generate_migration_meta(app_config=app_config)

    if auto:
        alter_statements = await AutoMigrationManager(
            auto_input=auto_input
        ).get_alter_statements(app_config=app_config)

        _alter_statements = list(
            chain(*[i.statements for i in alter_statements])
        )
        extra_imports = sorted(
            set(chain(*[i.extra_imports for i in alter_statements])),
            key=lambda x: x.__repr__(),
        )
        extra_definitions = sorted(
            set(chain(*[i.extra_definitions for i in alter_statements])),
        )

        if sum(len(i.statements) for i in alter_statements) == 0:
            raise NoChanges()

        file_contents = render_template(
            migration_id=meta.migration_id,
            auto=True,
            alter_statements=_alter_statements,
            extra_imports=extra_imports,
            extra_definitions=extra_definitions,
            app_name=app_config.app_name,
            description=description,
        )
    else:
        file_contents = render_template(
            migration_id=meta.migration_id, auto=False, description=description
        )

    # Beautify the file contents a bit.
    file_contents = black.format_str(
        file_contents, mode=black.FileMode(line_length=82)
    )

    with open(meta.migration_path, "w") as f:
        f.write(file_contents)

    return meta


###############################################################################


class AutoMigrationManager(BaseMigrationManager):
    def __init__(self, auto_input: Optional[str] = None, *args, **kwargs):
        self.auto_input = auto_input
        super().__init__(*args, **kwargs)

    async def get_alter_statements(
        self, app_config: AppConfig
    ) -> list[AlterStatements]:
        """
        Works out which alter statements are required.
        """
        migration_managers = await self.get_migration_managers(
            app_config=app_config
        )

        schema_snapshot = SchemaSnapshot(migration_managers)
        snapshot = schema_snapshot.get_snapshot()

        # Now get the current schema:
        current_diffable_tables = [
            DiffableTable(
                class_name=i.__name__,
                tablename=i._meta.tablename,
                columns=i._meta.non_default_columns,
                schema=i._meta.schema,
            )
            for i in app_config.table_classes
        ]

        # Compare the current schema with the snapshot
        differ = SchemaDiffer(
            schema=current_diffable_tables,
            schema_snapshot=snapshot,
            auto_input=self.auto_input,
        )
        return differ.get_alter_statements()


###############################################################################


async def new(
    app_name: str,
    auto: bool = False,
    desc: str = "",
    auto_input: Optional[str] = None,
):
    """
    Creates a new migration file in the migrations folder.

    :param app_name:
        The app to create a migration for. Specify a value of 'all' to create
        migrations for all apps (use in conjunction with --auto).
    :param auto:
        Auto create the migration contents.
    :param desc:
        A description of what the migration does, for example --desc='adding
        name column'.
    :param auto_input:
        If provided, all prompts for user input will automatically have this
        entered. For example, --auto_input='y'.

    """
    engine = Finder().get_engine()
    if auto and isinstance(engine, SQLiteEngine):
        colored_warning("Auto migrations aren't fully supported by SQLite.")

    if app_name == "all" and not auto:
        raise ValueError(
            "Only use `--app_name=all` in conjunction with `--auto`."
        )

    app_names = (
        sorted(
            BaseMigrationManager().get_app_names(
                sort_by_migration_dependencies=False
            )
        )
        if app_name == "all"
        else [app_name]
    )

    for app_name in app_names:
        print_heading(app_name)
        print("🚀 Creating new migration ...")

        app_config = Finder().get_app_config(app_name=app_name)

        _create_migrations_folder(app_config.resolved_migrations_folder_path)

        try:
            await _create_new_migration(
                app_config=app_config,
                auto=auto,
                description=desc,
                auto_input=auto_input,
            )
        except NoChanges:
            print("🏁 No changes detected.")

    print("\n✅ Finished\n")



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional, Union

from piccolo.querystring import QueryString
from piccolo.utils.encoding import dump_json

if TYPE_CHECKING:
    from piccolo.columns.column_types import JSON


class JSONQueryString(QueryString):

    def clean_value(self, value: Any):
        if not isinstance(value, (str, QueryString)):
            value = dump_json(value)
        return value

    def __eq__(self, value) -> QueryString:  # type: ignore[override]
        value = self.clean_value(value)
        return QueryString("{} = {}", self, value)

    def __ne__(self, value) -> QueryString:  # type: ignore[override]
        value = self.clean_value(value)
        return QueryString("{} != {}", self, value)

    def eq(self, value) -> QueryString:
        return self.__eq__(value)

    def ne(self, value) -> QueryString:
        return self.__ne__(value)


class GetChildElement(JSONQueryString):
    """
    Allows you to get a child element from a JSON object.

    You can access this via the ``arrow`` function on ``JSON`` and ``JSONB``
    columns.

    """

    def __init__(
        self,
        identifier: Union[JSON, QueryString],
        key: Union[str, int, QueryString],
        alias: Optional[str] = None,
    ):
        if isinstance(key, int):
            # asyncpg only accepts integer keys if we explicitly mark it as an
            # int.
            key = QueryString("{}::int", key)

        super().__init__("{} -> {}", identifier, key, alias=alias)

    def arrow(self, key: Union[str, int, QueryString]) -> GetChildElement:
        """
        This allows you to drill multiple levels deep into a JSON object if
        needed.

        For example::

            >>> await RecordingStudio.select(
            ...     RecordingStudio.name,
            ...     RecordingStudio.facilities.arrow(
            ...         "instruments"
            ...     ).arrow(
            ...         "drum_kits"
            ...     ).as_alias("drum_kits")
            ... ).output(load_json=True)
            [
                {'name': 'Abbey Road', 'drum_kits': 2},
                {'name': 'Electric Lady', 'drum_kits': 3}
            ]

        """
        return GetChildElement(identifier=self, key=key, alias=self._alias)

    def __getitem__(
        self, value: Union[str, int, QueryString]
    ) -> GetChildElement:
        return GetChildElement(identifier=self, key=value, alias=self._alias)


class GetElementFromPath(JSONQueryString):
    """
    Allows you to retrieve an element from a JSON object by specifying a path.
    It can be several levels deep.

    You can access this via the ``from_path`` function on ``JSON`` and
    ``JSONB`` columns.

    """

    def __init__(
        self,
        identifier: Union[JSON, QueryString],
        path: list[Union[str, int]],
        alias: Optional[str] = None,
    ):
        """
        :param path:
            For example: ``["technician", 0, "name"]``.

        """
        super().__init__(
            "{} #> {}",
            identifier,
            [str(i) if isinstance(i, int) else i for i in path],
            alias=alias,
        )



from __future__ import annotations

import datetime
from collections.abc import Callable
from enum import Enum
from typing import Union

from .timestamp import TimestampCustom, TimestampNow, TimestampOffset


class TimestamptzOffset(TimestampOffset):
    @property
    def cockroach(self):
        interval_string = self.get_postgres_interval_string(
            ["days", "hours", "minutes", "seconds"]
        )
        return f"CURRENT_TIMESTAMP + INTERVAL '{interval_string}'"

    def python(self):
        return datetime.datetime.now(
            tz=datetime.timezone.utc
        ) + datetime.timedelta(
            days=self.days,
            hours=self.hours,
            minutes=self.minutes,
            seconds=self.seconds,
        )


class TimestamptzNow(TimestampNow):
    """
    The current timestamp in UTC.
    """

    @property
    def cockroach(self):
        return "current_timestamp"

    def python(self):
        return datetime.datetime.now(tz=datetime.timezone.utc)


class TimestamptzCustom(TimestampCustom):
    @property
    def cockroach(self):
        return "'{}'".format(self.datetime.isoformat().replace("T", " "))

    @property
    def datetime(self):
        return datetime.datetime(
            year=self.year,
            month=self.month,
            day=self.day,
            hour=self.hour,
            minute=self.minute,
            second=self.second,
            microsecond=self.microsecond,
            tzinfo=datetime.timezone.utc,
        )

    @classmethod
    def from_datetime(cls, instance: datetime.datetime):  # type: ignore
        if instance.tzinfo is not None:
            instance = instance.astimezone(datetime.timezone.utc)
        return cls(
            year=instance.year,
            month=instance.month,
            day=instance.day,
            hour=instance.hour,
            minute=instance.minute,
            second=instance.second,
            microsecond=instance.microsecond,
        )


TimestamptzArg = Union[
    TimestamptzCustom,
    TimestamptzNow,
    TimestamptzOffset,
    Enum,
    None,
    datetime.datetime,
    Callable[[], datetime.datetime],
]


__all__ = [
    "TimestamptzArg",
    "TimestamptzCustom",
    "TimestamptzNow",
    "TimestamptzOffset",
]
